"""Functions for changing the installed version of Ladybug Tools."""
import os
import json
import subprocess

try:
    from ladybug.config import folders
    from ladybug.futil import nukedir, copy_file_tree, \
        download_file_by_name, unzip_file
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

from ..config import folders as lbr_folders
from ..pythonpath import iron_python_search_path, create_python_package_dir

# find the location where the Grasshopper user objects are stored
UO_DIRECTORY = lbr_folders.uo_folder
GHA_DIRECTORY = lbr_folders.gha_folder


def get_gem_directory():
    """Get the directory where measures distributed with Ladybug Tools are installed."""
    measure_folder = os.path.join(folders.ladybug_tools_folder, 'resources', 'measures')
    if not os.path.isdir(measure_folder):
        os.makedirs(measure_folder)
    return measure_folder


def get_standards_directory():
    """Get the directory where Honeybee standards are installed."""
    hb_folder = os.path.join(folders.ladybug_tools_folder, 'resources', 'standards')
    if not os.path.isdir(hb_folder):
        os.makedirs(hb_folder)
    return hb_folder


def remove_dist_info_files(directory):
    """Remove all of the PyPI .dist-info folders from a given directory.

    Args:
        directory: A directory containing .dist-info folders to delete.
    """
    for fold in os.listdir(directory):
        if fold.endswith('.dist-info'):
            nukedir(os.path.join(directory, fold), rmdir=True)


def get_config_dict():
    """Get a dictionary of the ladybug configurations.

    This is needed in order to put the configurations back after update.
    """
    with open(folders.config_file, 'r') as cfg:
        config_dict = json.load(cfg)
    return config_dict


def set_config_dict(config_dict):
    """Set the configurations using a dictionary.

    Args:
        config_dict: A dictionary of configuration paths.
    """
    with open(folders.config_file, 'w') as fp:
        json.dump(config_dict, fp, indent=4)


def update_libraries_pip(python_exe, package_name, version=None, target=None):
    """Change python libraries to be of a specific version using pip.

    Args:
        python_exe: The path to the Python executable to be used for installation.
        package_name: The name of the PyPI package to install.
        version: An optional string for the version of the package to install.
            If None, the library will be updated to the latest version with -U.
        target: An optional target directory into which the package will be installed.
        """
    # build up the command using the inputs
    if version is not None:
        package_name = '{}=={}'.format(package_name, version)
    cmds = [python_exe, '-m', 'pip', 'install', package_name]
    if version is None:
        cmds.append('-U')
    if target is not None:
        cmds.extend(['--target', target, '--upgrade'])

    # execute the command and print any errors
    print('Installing "{}" version via pip'.format(package_name))
    use_shell = True if os.name == 'nt' else False
    process = subprocess.Popen(
        cmds, shell=use_shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output = process.communicate()
    stdout, stderr = output
    error_msg = 'Package "{}" may not have been updated correctly\n' \
        'or its usage in the plugin may have changed. See pip stderr below:\n' \
        '{}'.format(package_name, stderr)
    return error_msg


def download_repo_github(repo, target_directory, version=None):
    """Download a repo of a particular version from from github.

    Args:
        repo: The name of a repo to be downloaded (eg. 'lbt-grasshopper').
        target_directory: The directory where the library should be downloaded.
        version: The version of the repository to download. If None, the most
            recent version will be downloaded. (Default: None)
    """
    # download files
    if version is None:
        url = 'https://github.com/ladybug-tools/{}/archive/master.zip'.format(repo)
    else:
        url = 'https://github.com/ladybug-tools/{}/archive/v{}.zip'.format(repo, version)
    zip_file = os.path.join(target_directory, '%s.zip' % repo)
    print('Downloading "{}"  github repository to: {}'.format(repo, target_directory))
    try:
        download_file_by_name(url, target_directory, zip_file)
    except ValueError:
        msg = 'Access is denied to: {}\nMake sure that you are running this command as' \
            ' an Administrator by right-clicking on\nRhino and selecting "Run As ' \
            'Administrator" before opening Grasshopper and\n running this ' \
            'component.'.format(target_directory)
        print(msg)
        raise ValueError(msg)

    # unzip the file
    unzip_file(zip_file, target_directory)

    # try to clean up the downloaded zip file
    try:
        os.remove(zip_file)
    except Exception:
        print('Failed to remove downloaded zip file: {}.'.format(zip_file))

    # return the directory where the unzipped files live
    if version is None:
        return os.path.join(target_directory, '{}-master'.format(repo))
    else:
        return os.path.join(target_directory, '{}-{}'.format(repo, version))


def latest_github_version(repo, target_directory):
    """Get the latest version tag of a particular repo on github.

    Args:
        repo: The name of a repo to be downloaded (eg. 'lbt-grasshopper').
        target_directory: The directory where the HTML Tags page should be downloaded.
    """
    # download the latest page with all of the tags
    url = 'https://github.com/ladybug-tools/{}/tags'.format(repo)
    tag_file = os.path.join(target_directory, '%s.html' % repo)
    try:
        download_file_by_name(url, target_directory, tag_file)
    except ValueError:
        msg = 'Access is denied to: {}\nMake sure that you are running this command as' \
            ' an Administrator.'.format(target_directory)
        print(msg)
        raise ValueError(msg)

    # parse the page to find the correct tag
    version_str, tag_found = None, False
    try:
        with open(tag_file) as tf:
            for row in tf:
                if 'data-test-selector="tag-title"' in row:
                    tag_found = True
                elif tag_found and '<a href' not in row:
                    version_str = row.strip().replace('v', '')
                    break
    except UnicodeDecodeError:
        pass  # machine lacks the encoding to read the HTML

    # try to clean up the downloaded html file
    try:
        os.remove(tag_file)
    except Exception:
        print('Failed to remove downloaded HTML file: {}.'.format(tag_file))
    return version_str


def update_requirements_version(uo_folder, lbt_version):
    """Update the version of lbt_grasshopper in the user object requirements.txt file.

    Args:
        uo_folder: The directory where the user objects currently exist.
        lbt_version: The version of LBT-grasshopper to be updated in the
            requirements.txt file.
    """
    req_file = os.path.join(uo_folder, 'requirements.txt')
    all_rows = []
    if os.path.isfile(req_file):
        with open(req_file) as rf:
            for row in rf:
                if row.startswith('lbt-grasshopper=='):
                    all_rows.append('lbt-grasshopper=={}\n'.format(lbt_version))
                else:
                    all_rows.append(row)
    else:
        all_rows.append('lbt-grasshopper=={}\n'.format(lbt_version))
    with open(req_file, 'w') as rf:
        rf.write(''.join(all_rows))


def parse_lbt_gh_versions(lbt_gh_folder):
    """Parse versions of compatible libs from a clone of the lbt-grasshopper repo.

    Args:
        lbt_gh_folder: Path to the clone of the lbt-grasshopper repo

    Returns:
        A dictionary of library versions formatted like so (but with actual version
        numbers in place of '0.0.0').

        {
            'lbt-dragonfly' = '0.0.0',
            'ladybug-rhino' = '0.0.0',
            'lbt-recipes' = '0.0.0',
            'honeybee-openstudio-gem' = '0.0.0',
            'lbt-measures' = '0.0.0',
            'honeybee-standards' = '0.0.0',
            'honeybee-energy-standards' = '0.0.0',
            'ladybug-grasshopper': '0.0.0',
            'honeybee-grasshopper-core': '0.0.0',
            'honeybee-grasshopper-radiance': '0.0.0',
            'honeybee-grasshopper-energy': '0.0.0',
            'dragonfly-grasshopper': '0.0.0',
            'ladybug-grasshopper-dotnet': '0.0.0'
        }
    """
    # set the names of the libraries to collect and the version dict
    version_dict = {
        'lbt-dragonfly': None,
        'ladybug-rhino': None,
        'lbt-recipes': None,
        'honeybee-standards': None,
        'honeybee-energy-standards': None,
        'honeybee-openstudio-gem': None,
        'lbt-measures': None,
        'ladybug-grasshopper': None,
        'honeybee-grasshopper-core': None,
        'honeybee-grasshopper-radiance': None,
        'honeybee-grasshopper-energy': None,
        'dragonfly-grasshopper': None,
        'ladybug-grasshopper-dotnet': None
    }
    libs_to_collect = list(version_dict.keys())

    def search_versions(version_file):
        """Search for version numbers within a .txt file."""
        with open(version_file) as ver_file:
            for row in ver_file:
                try:
                    library, version = row.strip().split('==')
                    if library in libs_to_collect:
                        version_dict[library] = version
                except Exception:  # not a row with a ladybug tools library
                    pass

    # search files for versions
    requirements = os.path.join(lbt_gh_folder, 'requirements.txt')
    dev_requirements = os.path.join(lbt_gh_folder, 'dev-requirements.txt')
    ruby_requirements = os.path.join(lbt_gh_folder, 'ruby-requirements.txt')
    search_versions(requirements)
    search_versions(dev_requirements)
    search_versions(ruby_requirements)
    return version_dict


def change_installed_version(version_to_install=None):
    """Change the currently installed version of Ladybug Tools.

    This requires an internet connection and will update all core libraries and
    Grasshopper components to the specified version_to_install.

    Args:
        version_to_install: An optional text string for the version of the LBT
            plugin to be installed. The input should contain only integers separated
            by two periods (eg. 1.0.0). If None, the Ladybug Tools plugin shall
            be updated to the latest available version. The version specified here
            does not need to be newer than the current installation and can be older
            but grasshopper plugin versions less than 0.3.0 are not supported.
            A list of all versions of the Grasshopper plugin can be found
            here - https://github.com/ladybug-tools/lbt-grasshopper/releases
    """
    # ensure that Python has been installed in the ladybug_tools folder
    py_exe, py_lib = folders.python_exe_path, folders.python_package_path
    assert py_exe is not None and py_lib is not None, \
        'No valid Python installation was found at: {}.\nThis is a requirement in ' \
        'order to continue with installation'.format(
            os.path.join(folders.ladybug_tools_folder, 'python'))

    # get the compatible versions of all the dependencies
    temp_folder = os.path.join(folders.ladybug_tools_folder, 'temp')
    lbt_gh_folder = download_repo_github(
        'lbt-grasshopper', temp_folder, version_to_install)
    ver_dict = parse_lbt_gh_versions(lbt_gh_folder)
    if version_to_install is None:
        version_to_install = latest_github_version('lbt-grasshopper', temp_folder)
    ver_dict['lbt-grasshopper'] = version_to_install

    # install the core libraries
    print('Installing Ladybug Tools core Python libraries.')
    config_dict = get_config_dict()  # load configs so they can be put back after update
    df_ver = ver_dict['lbt-dragonfly']
    stderr = update_libraries_pip(py_exe, 'lbt-dragonfly', df_ver)
    if os.path.isdir(os.path.join(py_lib, 'lbt_dragonfly-{}.dist-info'.format(df_ver))):
        print('Ladybug Tools core Python libraries successfully installed!\n ')
    else:
        print(stderr)
    set_config_dict(config_dict)  # restore the previous configurations

    # install the lbt_recipes package
    print('Installing Ladybug Tools Recipes.')
    rec_ver = ver_dict['lbt-recipes']
    stderr = update_libraries_pip(py_exe, 'lbt-recipes', rec_ver)
    if os.path.isdir(os.path.join(py_lib, 'lbt_recipes-{}.dist-info'.format(rec_ver))):
        print('Ladybug Tools Recipes successfully installed!\n ')
    else:
        print(stderr)

    # install the library needed for interaction with Rhino
    print('Installing ladybug-rhino Python library.')
    rh_ver = ver_dict['ladybug-rhino']
    stderr = update_libraries_pip(py_exe, 'ladybug-rhino', rh_ver)
    if os.path.isdir(os.path.join(py_lib, 'ladybug_rhino-{}.dist-info'.format(rh_ver))):
        print('Ladybug-rhino Python library successfully installed!\n ')
    else:
        print(stderr)

    # make sure libraries are copied to the rhino scripts folder on Mac
    if os.name != 'nt':
        iron_python_search_path(create_python_package_dir())

    # install the grasshopper components
    print('Installing Ladybug Tools Grasshopper components.')
    gh_ver = ver_dict['lbt-grasshopper']
    stderr = update_libraries_pip(py_exe, 'lbt-grasshopper', gh_ver, UO_DIRECTORY)
    lb_gh_ver = ver_dict['ladybug-grasshopper']
    lb_gh_info = 'ladybug_grasshopper-{}.dist-info'.format(lb_gh_ver)
    if os.path.isdir(os.path.join(UO_DIRECTORY, lb_gh_info)):
        print('Ladybug Tools Grasshopper components successfully installed!\n ')
        remove_dist_info_files(UO_DIRECTORY)  # remove the .dist-info files
    else:
        print(stderr)
    if version_to_install is not None:
        update_requirements_version(UO_DIRECTORY, version_to_install)

    # install the .gha Grasshopper components
    gha_location = os.path.join(GHA_DIRECTORY, 'ladybug_grasshopper_dotnet')
    if os.path.isdir(gha_location):
        msg = '.gha files already exist in your Components folder and cannot be ' \
            'deleted while Grasshopper is open.\nThese .gha files rarely change ' \
            'and so it is not critical that they be updated. However, if you ' \
            'want to be sure that you have the latest version installed, then close ' \
            'Grasshopper, delete the ladybug_grasshopper_dotnet folder at\n{}\nand ' \
            'rerun this versioner component to get the new .gha files.\n'.format(
                gha_location)
        print(msg)
    else:
        gha_ver = ver_dict['ladybug-grasshopper-dotnet']
        stderr = update_libraries_pip(
            py_exe, 'ladybug-grasshopper-dotnet', gha_ver, GHA_DIRECTORY)
        package_dir = os.path.join(
            GHA_DIRECTORY, 'ladybug_grasshopper_dotnet-{}.dist-info'.format(gha_ver))
        if os.path.isdir(package_dir):
            print('Ladybug Tools .gha Grasshopper components successfully installed!\n ')
            remove_dist_info_files(GHA_DIRECTORY)  # remove the dist-info files
        else:
            print(stderr)

    # install the honeybee-openstudio ruby gem
    gem_ver = ver_dict['honeybee-openstudio-gem']
    print('Installing Honeybee-OpenStudio gem version {}.'.format(gem_ver))
    gem_dir = get_gem_directory()
    base_folder = download_repo_github('honeybee-openstudio-gem', gem_dir, gem_ver)
    source_folder = os.path.join(base_folder, 'lib')
    lib_folder = os.path.join(gem_dir, 'honeybee_openstudio_gem', 'lib')
    print('Copying "honeybee_openstudio_gem" source code to {}\n '.format(lib_folder))
    copy_file_tree(source_folder, lib_folder)
    nukedir(base_folder, True)

    # install the lbt-measures ruby gem
    mea_ver = ver_dict['lbt-measures']
    print('Installing Ladybug Tools Measures version {}.'.format(mea_ver))
    base_folder = download_repo_github('lbt-measures', gem_dir, mea_ver)
    source_folder = os.path.join(base_folder, 'lib')
    print('Copying "lbt_measures" source code to {}\n '.format(gem_dir))
    copy_file_tree(source_folder, gem_dir)
    nukedir(base_folder, True)

    # always update the honeybee-energy-standards package
    print('Installing Honeybee energy standards.')
    stand_dir = get_standards_directory()
    hes_ver = ver_dict['honeybee-energy-standards']
    if os.path.isdir(os.path.join(stand_dir, 'honeybee_energy_standards')):
        nukedir(os.path.join(stand_dir, 'honeybee_energy_standards'), True)
    stderr = update_libraries_pip(
        py_exe, 'honeybee-energy-standards', hes_ver, stand_dir)
    hes_info = 'honeybee_energy_standards-{}.dist-info'.format(hes_ver)
    if os.path.isdir(os.path.join(stand_dir, hes_info)):
        print('Honeybee energy standards successfully installed!\n ')
        remove_dist_info_files(stand_dir)  # remove the dist-info files
    else:
        print(stderr)

    # delete the temp folder and give a completion message
    nukedir(temp_folder, True)
    version = 'LATEST' if version_to_install is None else version_to_install
    success_msg = 'Change to Version {} Successful!'.format(version)
    restart_msg = 'RESTART RHINO to load the new components + library.'
    sync_msg = 'The "LB Sync Grasshopper File" component can be used\n' \
        'to sync Grasshopper definitions with your new installation.'
    for msg in (success_msg, restart_msg, sync_msg):
        print(msg)



"""Collection of methods for downloading files securely using .NET libraries."""
import os

try:
    from ladybug.futil import preparedir
except ImportError as e:
    raise ImportError("Failed to import ladybug.\n{}".format(e))

try:
    import System.Net
except ImportError as e:
    print("Failed to import Windows/.NET libraries\n{}".format(e))


def download_file_by_name(url, target_folder, file_name, mkdir=False):
    """Download a file to a directory.

    Args:
        url: A string to a valid URL.
        target_folder: Target folder for download (e.g. c:/ladybug)
        file_name: File name (e.g. testPts.zip).
        mkdir: Set to True to create the directory if doesn't exist (Default: False)
    """
    # create the target directory.
    if not os.path.isdir(target_folder):
        if mkdir:
            preparedir(target_folder)
        else:
            created = preparedir(target_folder, False)
            if not created:
                raise ValueError("Failed to find %s." % target_folder)
    file_path = os.path.join(target_folder, file_name)

    # set the security protocol to the most recent version
    try:
        # TLS 1.2 is needed to download over https
        System.Net.ServicePointManager.SecurityProtocol = \
            System.Net.SecurityProtocolType.Tls12
    except AttributeError:
        # TLS 1.2 is not provided by MacOS .NET in Rhino 5
        if url.lower().startswith('https'):
            print('This system lacks the necessary security'
                  ' libraries to download over https.')

    # attempt to download the file
    client = System.Net.WebClient()
    try:
        client.DownloadFile(url, file_path)
    except Exception as e:
        raise Exception(' Download failed with the error:\n{}'.format(e))


def download_file(url, file_path, mkdir=False):
    """Write a string of data to file.

    Args:
        url: A string to a valid URL.
        file_path: Full path to intended download location (e.g. c:/ladybug/testPts.pts)
        mkdir: Set to True to create the directory if doesn't exist (Default: False)
    """
    folder, fname = os.path.split(file_path)
    return download_file_by_name(url, folder, fname, mkdir)



"""Functions to add text to the Rhino scene and create Grasshopper text objects."""
from __future__ import division
import math

from .fromgeometry import from_plane
from .color import black

try:
    import System.Guid as guid
except ImportError as e:
    print("Failed to import System\n{}".format(e))

try:
    import Rhino as rh
except ImportError as e:
    raise ImportError("Failed to import Rhino.\n{}".format(e))

try:
    import Grasshopper as gh
except ImportError:
    print('Failed to import Grasshopper.\n'
          'Only functions for adding text to Rhino will be available.')


def text_objects(text, plane, height, font='Arial',
                 horizontal_alignment=0, vertical_alignment=5):
    """Generate a Bake-able Grasshopper text object from a text string and ladybug Plane.

    Args:
        text: A text string to be converted to a a Grasshopper text object.
        plane: A Ladybug Plane object to locate and orient the text in the Rhino scene.
        height: A number for the height of the text in the Rhino scene.
        font: An optional text string for the font in which to draw the text.
        horizontal_alignment: An optional integer to specify the horizontal alignment
             of the text. Choose from: (0 = Left, 1 = Center, 2 = Right)
        vertical_alignment: An optional integer to specify the vertical alignment
             of the text. Choose from: (0 = Top, 1 = MiddleOfTop, 2 = BottomOfTop,
             3 = Middle, 4 = MiddleOfBottom, 5 = Bottom, 6 = BottomOfBoundingBox)
    """
    txt = rh.Display.Text3d(text, from_plane(plane), height)
    txt.FontFace = font
    txt.HorizontalAlignment = AlignmentTypes.horizontal(horizontal_alignment)
    txt.VerticalAlignment = AlignmentTypes.vertical(vertical_alignment)
    return TextGoo(txt)


"""____________EXTRA HELPER OBJECTS____________"""


class TextGoo(gh.Kernel.Types.GH_GeometricGoo[rh.Display.Text3d],
              gh.Kernel.IGH_BakeAwareData, gh.Kernel.IGH_PreviewData):
    """A Text object that can be baked and transformed in Grasshopper.

    The code for this entire class was taken from David Rutten and Giulio Piacentino's
    script described here:
    https://discourse.mcneel.com/t/creating-text-objects-and-outputting-them-as-normal-rhino-geometry/47834/7

    Args:
        text: A Rhino Text3d object.
    """

    def __init__(self, text):
        """Initialize Bake-able text."""
        self.m_value = text

    @staticmethod
    def DuplicateText3d(original):
        if original is None:
            return None
        text = rh.Display.Text3d(original.Text, original.TextPlane, original.Height)
        text.Bold = original.Bold
        text.Italic = original.Italic
        text.FontFace = original.FontFace
        return text

    def DuplicateGeometry(self):
        return TextGoo(TextGoo.DuplicateText3d(self.m_value))

    def get_TypeName(self):
        return "3D Text"

    def get_TypeDescription(self):
        return "3D Text"

    def get_Boundingbox(self):
        if self.m_value is None:
            return rh.Geometry.BoundingBox.Empty
        return self.m_value.BoundingBox

    def GetBoundingBox(self, xform):
        if self.m_value is None:
            return rh.Geometry.BoundingBox.Empty
        box = self.m_value.BoundingBox
        corners = xform.TransformList(box.GetCorners())
        return rh.Geometry.BoundingBox(corners)

    def Transform(self, xform):
        text = TextGoo.DuplicateText3d(self.m_value)
        if text is None:
            return TextGoo(None)

        plane = text.TextPlane
        point = plane.PointAt(1, 1)

        plane.Transform(xform)
        point.Transform(xform)
        dd = point.DistanceTo(plane.Origin)

        text.TextPlane = plane
        text.Height *= dd / math.sqrt(2)
        new_text = TextGoo(text)

        new_text.m_value.Bold = self.m_value.Bold
        new_text.m_value.Italic = self.m_value.Italic
        new_text.m_value.FontFace = self.m_value.FontFace
        return new_text

    def Morph(self, xmorph):
        return self.DuplicateGeometry()

    def get_ClippingBox(self):
        return self.get_Boundingbox()

    def DrawViewportWires(self, args):
        if self.m_value is None:
            return
        color = black() if black is not None else args.Color
        args.Pipeline.Draw3dText(self.m_value, color)

    def DrawViewportMeshes(self, args):
        # Do not draw in meshing layer.
        pass

    def BakeGeometry(self, doc, att, id):
        id = guid.Empty
        if self.m_value is None:
            return False, id
        if att is None:
            att = doc.CreateDefaultAttributes()
        original_plane = None
        d_txt = self.m_value.Text
        nl_count = len(d_txt.split('\n')) - 1
        if nl_count > 1 and str(self.m_value.VerticalAlignment) == 'Bottom':
            y_ax = rh.Geometry.Vector3d(self.m_value.TextPlane.YAxis)
            txt_h = self.m_value.Height * (3 / 2)
            m_vec = rh.Geometry.Vector3d.Multiply(y_ax, txt_h * -nl_count)
            original_plane = self.m_value.TextPlane
            new_plane = rh.Geometry.Plane(self.m_value.TextPlane)
            new_plane.Translate(m_vec)
            self.m_value.TextPlane = new_plane
        self.m_value.Height = self.m_value.Height * (2 / 3)
        id = doc.Objects.AddText(self.m_value, att)
        self.m_value.Height = self.m_value.Height * (3 / 2)
        if original_plane is not None:
            self.m_value.TextPlane = original_plane
        return True, id

    def ScriptVariable(self):
        """Overwrite Grasshopper ScriptVariable method."""
        return self

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        if self.m_value is None:
            return "<null>"
        return self.m_value.Text


class AlignmentTypes(object):
    """Enumeration of text alignment types."""

    _HORIZONTAL = (rh.DocObjects.TextHorizontalAlignment.Left,
                   rh.DocObjects.TextHorizontalAlignment.Center,
                   rh.DocObjects.TextHorizontalAlignment.Right)

    _VERTICAL = (rh.DocObjects.TextVerticalAlignment.Top,
                 rh.DocObjects.TextVerticalAlignment.MiddleOfTop,
                 rh.DocObjects.TextVerticalAlignment.BottomOfTop,
                 rh.DocObjects.TextVerticalAlignment.Middle,
                 rh.DocObjects.TextVerticalAlignment.MiddleOfBottom,
                 rh.DocObjects.TextVerticalAlignment.Bottom,
                 rh.DocObjects.TextVerticalAlignment.BottomOfBoundingBox)

    @classmethod
    def horizontal(cls, field_number):
        """Get a Rhino horizontal alignment object by its integer field number.

        * 0 - Left
        * 1 - Center
        * 2 - Right

        """
        return cls._HORIZONTAL[field_number]

    @classmethod
    def vertical(cls, field_number):
        """Get a Rhino vertical alignment object by its integer field number.

        * 0 - Top
        * 1 - MiddleOfTop
        * 2 - BottomOfTop
        * 3 - Middle
        * 4 - MiddleOfBottom
        * 5 - Bottom
        * 6 - BottomOfBoundingBox

        """
        return cls._VERTICAL[field_number]



"""Command Line Interface (CLI) entry point for ladybug rhino.

Note:

    Do not import this module in your code directly. For running the commands,
    execute them from the command line or as a subprocess
    (e.g. ``subprocess.call(['ladybug-rhino', 'viz'])``)

Ladybug rhino is using click (https://click.palletsprojects.com/en/7.x/) for
creating the CLI.
"""

import sys
import os
import json
import logging
import click

try:
    from ladybug.config import folders as lb_folders
except ImportError as e:
    raise ImportError("Failed to import ladybug.\n{}".format(e))

from ladybug_rhino.config import folders
from ladybug_rhino.pythonpath import create_python_package_dir, iron_python_search_path
from ladybug_rhino.ghpath import copy_components_packages, \
    clean_userobjects, clean_libraries
from ladybug_rhino.resourcepath import setup_resource_folders
from ladybug_rhino.versioning.change import change_installed_version

_logger = logging.getLogger(__name__)


@click.group()
@click.version_option()
def main():
    pass


@main.command('viz')
def viz():
    """Check if ladybug_rhino is flying!"""
    click.echo('viiiiiiiiiiiiizzzzzzzzz!')


@main.command('setup-user-environment')
@click.option(
    '--component-directory', default=None, help='The path to a directory that '
    'contains all of the Ladybug Tools Grasshopper python packages to be copied '
    '(both user object packages and dotnet gha packages). If unspecified, this command '
    'will search for the site-packages folder in the ladybug_tools folder. If '
    'they are not found, no user objects will be copied.',
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True))
@click.option(
    '--python-package-dir', help='Path to the directory with the python '
    'packages, which will be added to the search path. If unspecified, this command '
    'will search for the site-packages folder in the ladybug_tools folder.',
    type=str, default=None, show_default=True)
@click.option(
    '--setup-resources/--overwrite-resources', ' /-o', help='Flag to note '
    'whether the user resources should be overwritten or they should only '
    'be set up if they do not exist, in which case existing resources will '
    'be preserved.', default=True)
def setup_user_environment(component_directory, python_package_dir, setup_resources):
    """Set up the entire environment for the current user.

    This includes setting the IronPython search path, copying the components to
    the user-specific folders, and creating any user resource folders if they
    do not already exist. Note that setting the IronPython search path won't
    work well if Rhino is open while running this command.
    """
    try:
        # set the ironpython search path
        if python_package_dir is None:
            python_package_dir = create_python_package_dir()
        if python_package_dir is not None:
            click.echo('Setting Rhino IronPython search path ...')
            new_settings = iron_python_search_path(python_package_dir, None)
            click.echo('Congratulations! Setting the search path in the following '
                       'file was successful:\n{}'.format('\n'.join(new_settings)))
        # copy the components if they exist
        if component_directory is None:
            component_directory = \
                os.path.join(lb_folders.ladybug_tools_folder, 'grasshopper')
        if os.path.isdir(component_directory):
            click.echo('Copying Grasshopper Components ...')
            copy_components_packages(component_directory)
            click.echo('Congratulations! All component packages are copied!')
        # set the user resources
        click.echo('Setting user-specific resources ...')
        overwrite = not setup_resources
        resource_folder = setup_resource_folders(overwrite)
        click.echo('Setting up user resources in the following '
                   'folder was successful:\n{}'.format(resource_folder))
    except Exception as e:
        _logger.exception('Setting up the user environment failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@main.command('set-python-search')
@click.option('--python-package-dir', help='Path to the directory with the python '
              'packages, which will be added to the search path. If None, this command '
              'will search for the site-packages folder in the ladybug_tools folder',
              type=str, default=None, show_default=True)
@click.option('--settings-file', help='Path to the Rhino settings file to which the '
              'python-package-dir will be added. If None, this command will search '
              'the current user folder for all copies of this file for the installed '
              'Rhino versions. ', type=str, default=None, show_default=True)
def set_python_search(python_package_dir, settings_file):
    """Set Rhino to search for libraries in a given directory."""
    try:
        # search for the python package directory if it is not there
        if python_package_dir is None:
            python_package_dir = create_python_package_dir()

        # set the search path
        click.echo('Setting Rhino IronPython search path ...')
        new_settings = iron_python_search_path(python_package_dir, settings_file)
        click.echo('Congratulations! Setting the search path in the following '
                   'file was successful:\n{}'.format('\n'.join(new_settings)))
    except Exception as e:
        _logger.exception('Setting IronPython search path failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@main.command('copy-gh-components')
@click.argument('component-directory', type=click.Path(
    exists=True, file_okay=False, dir_okay=True, resolve_path=True))
def copy_gh_components(component_directory):
    """Copy all component packages to the UserObjects and Libraries folder.

    \b
    Args:
        component_directory: The path to a directory that contains all of the Ladybug
            Tools Grasshopper python packages to be copied (both user object
            packages and dotnet gha packages).
    """
    try:
        # copy the grasshopper components
        click.echo('Copying Grasshopper Components ...')
        copy_components_packages(component_directory)
        click.echo('Congratulations! All component packages are copied!')
    except Exception as e:
        _logger.exception('Copying Grasshopper components failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@main.command('remove-gh-components')
def remove_gh_components():
    """Remove all component packages to the UserObjects and Libraries folder."""
    try:
        # copy the grasshopper components
        click.echo('Removing Grasshopper Components ...')
        clean_userobjects()
        clean_libraries()
        click.echo('Congratulations! All component packages are removed!')
    except Exception as e:
        _logger.exception('Removing Grasshopper components failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@main.command('setup-resources')
@click.option('--setup-only/--overwrite', ' /-o', help='Flag to note '
              'whether the user resources should be overwritten or they should only '
              'be set up if they do not exist, in which case existing resources will '
              'be preserved.', default=True)
def setup_resources(setup_only):
    """Set up user resource folders in their respective locations."""
    try:
        overwrite = not setup_only
        # setup the resource folders
        click.echo('Setting user-specific resources ...')
        resource_folder = setup_resource_folders(overwrite)
        click.echo('Setting up user resources in the following '
                   'folder was successful:\n{}'.format(resource_folder))
    except Exception as e:
        _logger.exception('Setting up resource folders failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@main.command('change-installed-version')
@click.option('--version', '-v', help=' An optional text string for the version of '
              'the LBT plugin to be installed. The input should contain only integers '
              'separated by two periods (eg. 1.0.0). If unspecified, the Ladybug '
              'Tools plugin shall be updated to the latest available version. The '
              'version specified here does not need to be newer than the current '
              'installation and can be older but grasshopper plugin versions less '
              'than 0.3.0 are not supported.', type=str, default=None)
def run_versioner_process(version):
    """Change the currently installed version of Ladybug Tools.

    This requires an internet connection and will update all core libraries and
    Grasshopper components to the specified version_to_install.
    """
    try:
        change_installed_version(version)
    except Exception as e:
        _logger.exception('Changing the installed version failed.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


@main.command('config')
@click.option('--output-file', help='Optional file to output the JSON string of '
              'the config object. By default, it will be printed out to stdout',
              type=click.File('w'), default='-', show_default=True)
def config_variables(output_file):
    """Get a JSON object with all configuration information."""
    try:
        config_dict = {
            'uo_folder': folders.uo_folder,
            'gha_folder': folders.gha_folder,
            'lbt_grasshopper_version': folders.lbt_grasshopper_version_str
        }
        output_file.write(json.dumps(config_dict, indent=4))
    except Exception as e:
        _logger.exception('Failed to retrieve configurations.\n{}'.format(e))
        sys.exit(1)
    else:
        sys.exit(0)


if __name__ == "__main__":
    main()



"""Functions for managing components differences and syncing component versions."""
import os

try:
    import System.Drawing
except ImportError:
    raise ImportError("Failed to import System.")

try:
    import Grasshopper.Kernel as gh
except ImportError:
    raise ImportError("Failed to import Grasshopper.")

from ..grasshopper import give_warning
from .userobject import UO_FOLDER, FOLDER_MAP


# list of valid change tags for export
CHANGE_TAGS = ('fix', 'release', 'feat', 'perf', 'docs', 'ignore')


def validate_change_type(change_type):
    """Check that a change type is a valid tag."""
    assert change_type in CHANGE_TAGS, 'Invalid change_type "{}". Choose from ' \
        'the following:\n{}'.format(change_type, ' '.join(CHANGE_TAGS))
    return change_type


def current_userobject_version(component):
    """Get the current installed version of a component.

    Args:
        component: A Grasshopper Python component with the same name as an
            installed user object. If no matching user object is found, this
            method returns None.
    """
    # load component from the folder where it should be
    assert component.Category in FOLDER_MAP, \
        'Unknown category: %s. Add category to folder_dict.' % component.Category
    fp = os.path.join(UO_FOLDER, FOLDER_MAP[component.Category], 'user_objects',
                      '%s.ghuser' % component.Name)

    if os.path.isfile(fp):  # if the component was found, parse out the version
        uo = gh.GH_UserObject(fp).InstantiateObject()
        # uo.Message returns None so we have to find it the old school way!
        for lc, line in enumerate(uo.Code.split('\n')):
            if lc > 200:  # this should never happen for valid userobjects
                raise ValueError(
                    'Failed to find version from UserObject for %s' % uo.Name
                )
            if line.strip().startswith("ghenv.Component.Message"):
                return line.split("=")[1].strip()[1:-1]
    else:  # there is no currently installed version of this userobject
        return None


def parse_version(semver_str):
    """Parse semantic version string into (major, minor, patch) tuple.

    Args:
        semver_str: Text for a component version (eg. "1.0.1").
    """
    try:
        major, minor, patch = [int(d) for d in semver_str.strip().split('.')]
    except ValueError:
        raise ValueError(
            '\nInvalid version format: %s\nYou must follow major.minor.patch format '
            'with 3 integer values' % semver_str
        )
    return major, minor, patch


def validate_version(current_version, new_version, change_type):
    """Validate that a version tag conforms to currently-installed version difference.

    Args:
        current_version: Semantic version string for the currently installed version.
        new_version: Semantic version string for the new component version.
        change_type: Text tag for the change type (eg. "fix").
    """
    if current_version is None:
        # this is the first time that this component is created; give it a pass
        print('    New component. No change in version: %s' % current_version)
        return True

    x, y, z = parse_version(current_version)
    parse_version(new_version)  # just make sure the semantic version format is valid

    msg = '\nFor a \'%s\' the component version should change to %s not %s.' \
        '\nFix the version or select the correct change type and try again.'
    if change_type == 'ignore':
        valid_version = new_version
    elif change_type == 'fix':
        valid_version = '.'.join(str(i) for i in (x, y, z + 1))
    elif change_type == 'feat' or change_type == 'perf':
        valid_version = '.'.join(str(i) for i in (x, y + 1, 0))
    elif change_type == 'release':
        valid_version = '.'.join(str(i) for i in (x + 1, 0, 0))
    elif change_type == 'docs':
        valid_version = '.'.join(str(i) for i in (x, y, z))
    else:
        raise ValueError('Invalid change_type: %s' % change_type)

    assert valid_version == new_version, msg % (change_type, valid_version, new_version)

    if current_version == new_version:
        print('    No change in version: %s' % current_version)
    else:
        print('    Version is changed from %s to %s.' % (current_version, new_version))


def has_version_changed(user_object, component):
    """Check if the version of a component has changed from a corresponding user object.

    Note that this method only works for user objects that have been dropped on the
    canvas. For comparing the version with a user object that hasn't been loaded from
    the component server by dropping it on the canvas, the current_userobject_version
    method should be used.

    Args:
        user_object: A Grasshopper user object component instance.
        component: The Grasshopper component object for which the version is
            being checked.
    """
    return not user_object.Message == component.Message


def update_port(p1, p2):
    """Update one port based on another.

    Args:
        p1: The first port object, which is the one used for updating.
        p2: The second port object, which will be updated base on p1.
    """
    if hasattr(p1, 'TypeHint'):  # input
        if p1.Name != p2.Name:
            p2.NickName = p1.NickName
            p2.Name = p1.Name
        if p1.TypeHint.TypeName != p2.TypeHint.TypeName:
            p2.TypeHint = p1.TypeHint
        if str(p1.Access) != str(p2.Access):
            p2.Access = p1.Access
        return True
    else:  # output
        if p1.Name != p2.Name:
            p2.NickName = p1.NickName
            p2.Name = p1.Name
        return True


def update_ports(c1, c2):
    """Update all of the ports of one component based on another.

    Args:
        c1: The first component object, which is the one used for updating.
        c2: The second component object, which will be updated base on c1.
    """
    for i in range(c1.Params.Input.Count):
        if not update_port(c1.Params.Input[i], c2.Params.Input[i]):
            return True

    for i in range(c1.Params.Output.Count):
        if not update_port(c1.Params.Output[i], c2.Params.Output[i]):
            return True

    return False


def input_output_changed(user_object, component):
    """Check if any of inputs or outputs have changed between two components.

    Args:
        user_object: A Grasshopper user object component instance.
        component: The Grasshopper component object for which the version is
            being checked.
    """
    if user_object.Params.Input.Count != component.Params.Input.Count:
        return True
    elif user_object.Params.Output.Count != component.Params.Output.Count:
        return True

    return update_ports(user_object, component)


def insert_new_user_object(user_object, component, doc):
    """Insert a new user object next to an existing component in the Grasshopper doc.

    Args:
        user_object: A Grasshopper user object component instance.
        component: The outdated component where the userobject will be inserted
            next to.
        doc: The Grasshopper document object.
    """
    # use component to find the location
    x = component.Attributes.Pivot.X + 30
    y = component.Attributes.Pivot.Y - 20

    # insert the new one
    user_object.Attributes.Pivot = System.Drawing.PointF(x, y)
    doc.AddObject(user_object, False, 0)


def mark_component(doc, component, note=None):
    """Put a circular red group around a component and label it with a note.

    Args:
        doc: The Grasshopper document object.
        component: A Grasshopper component object on the canvas to be circled.
        note: Text for the message to be displayed on the circle. If None, a
            default message about input/output change wil be used.
    """
    note = note or 'There is a change in the input or output! ' \
        'Replace this component manually.'
    grp = gh.Special.GH_Group()
    grp.CreateAttributes()
    grp.Border = gh.Special.GH_GroupBorder.Blob
    grp.AddObject(component.InstanceGuid)
    grp.Colour = System.Drawing.Color.IndianRed  # way to pick a racist color name, .NET
    grp.NickName = note
    doc.AddObject(grp, False)
    return True


def sync_component(component, syncing_component):
    """Sync a component on the canvas with its corresponding installed version.

    This includes identifying if the component by name in the user object folder,
    injecting the code from that user object into the component, and (if the
    component inputs or outputs have changed) circling the component in red +
    dropping the new user object next to the component.

    Args:
        component: A Grasshopper component object on the canvas to be circled.
        syncing_component: An object for the component that is doing the syncing.
            This will be used to give warnings and access the Grasshopper doc.
            Typically, this can be accessed through the ghenv.Component call.
    """
    # identify the correct user object sub-folder to which the component belongs
    ghuser_file = '%s.ghuser' % component.Name
    if str(component.Name).startswith(('LB', 'HB', 'DF')):
        fp = os.path.join(UO_FOLDER, FOLDER_MAP[component.Category],
                          'user_objects', ghuser_file)
    elif str(component.Name).startswith(('Ladybug', 'Honeybee', 'HoneybeePlus')):
        category = str(component.Name).split('_')[0]
        fp = os.path.join(UO_FOLDER, category, ghuser_file)
    else:  # unidentified plugin; see if we can find it in the root
        fp = os.path.join(UO_FOLDER, ghuser_file)

    # check to see if the user object is installed
    if not os.path.isfile(fp):
        # check if it's a component with a name change
        alt_fp = fp.replace('Vizualize', 'Visualize')
        alt_fp = alt_fp.replace('gbXML', 'gbXML OSM IDF')
        if os.path.isfile(alt_fp):
            fp = alt_fp
        elif component.Category in FOLDER_MAP:  # see if category has a folder
            fp = os.path.join(UO_FOLDER, FOLDER_MAP[component.Category], ghuser_file)
        if not os.path.isfile(fp):  # see if the component is in the root
            fp = os.path.join(UO_FOLDER, ghuser_file)
            if not os.path.isfile(fp):  # all hope is lost; component not installed
                warning = 'Failed to find the userobject for %s' % component.Name
                give_warning(syncing_component, warning)
                return False

    # the the instance of the user object from the file
    uo = gh.GH_UserObject(fp).InstantiateObject()

    # check to see if the version of the userobject has changed
    if not has_version_changed(uo, component):
        return False

    # the version has changed, let's update the code
    component.Code = uo.Code
    doc = syncing_component.OnPingDocument()  # get the Grasshopper document

    # define the callback function and update the solution
    def call_back(document):
        component.ExpireSolution(False)

    doc.ScheduleSolution(2, gh.GH_Document.GH_ScheduleDelegate(call_back))

    # check if the inputs or outputs have changed
    if input_output_changed(uo, component):
        insert_new_user_object(uo, component, doc)
        mark_component(doc, component)  # mark component with a warning to the user
        return 'Cannot update %s. Replace manually.' % component.Name

    return 'Updated %s' % component.Name



"""Functions to add text to the Rhino scene and create Grasshopper text objects."""
from ladybug_display.altnumber import Default

from .color import color_to_color, argb_color_to_color
from .fromgeometry import from_plane
from .bakegeometry import _get_attributes, bake_point2d, bake_vector2d, bake_ray2d, \
    bake_linesegment2d, bake_arc2d, bake_polygon2d, bake_polyline2d, bake_mesh2d, \
    bake_point3d, bake_vector3d, bake_ray3d, bake_linesegment3d, bake_plane, \
    bake_arc3d, bake_polyline3d, bake_mesh3d, bake_face3d, bake_polyface3d, \
    bake_sphere, bake_cone, bake_cylinder

try:
    import Rhino.Display as rd
    import Rhino.DocObjects as docobj
    import Rhino.RhinoDoc as rhdoc
    doc = rhdoc.ActiveDoc
except ImportError as e:
    raise ImportError("Failed to import Rhino document attributes.\n{}".format(e))

TEXT_HORIZ = {
    'Left': docobj.TextHorizontalAlignment.Left,
    'Center': docobj.TextHorizontalAlignment.Center,
    'Right': docobj.TextHorizontalAlignment.Right
}
TEXT_VERT = {
    'Top': docobj.TextVerticalAlignment.Top,
    'Middle': docobj.TextVerticalAlignment.Middle,
    'Bottom': docobj.TextVerticalAlignment.Bottom
}
LINE_WIDTH_FACTOR = 3.779528  # factor to translate pixel width to millimeters
LINE_TYPES = {
    'Continuous': -1,
    'Dashed': -1,
    'Dotted': -1,
    'DashDot': -1
}
_display_lts = ('Continuous', 'Dashed', 'Dots', 'DashDot')
for i, lt in enumerate(doc.Linetypes):
    lt_name = lt.Name
    for dlt in _display_lts:
        if lt_name == dlt:
            dlt = dlt if dlt != 'Dots' else 'Dotted'
            LINE_TYPES[dlt] = i
            break


"""____________BAKE 2D DISPLAY GEOMETRY TO THE RHINO SCENE____________"""


def bake_display_vector2d(vector, z=0, layer_name=None, attributes=None):
    """Add DisplayVector2D to the Rhino scene as a Line with an Arrowhead."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, vector)
    return bake_vector2d(vector.geometry, z, attributes=attrib)


def bake_display_point2d(point, z=0, layer_name=None, attributes=None):
    """Add ladybug Point2D to the Rhino scene as a Point."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, point)
    return bake_point2d(point.geometry, z, attributes=attrib)


def bake_display_ray2d(ray, z=0, layer_name=None, attributes=None):
    """Add DisplayRay2D to the Rhino scene as a Line with an Arrowhead."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, ray)
    return bake_ray2d(ray.geometry, z, attributes=attrib)


def bake_display_linesegment2d(line, z=0, layer_name=None, attributes=None):
    """Add DisplayLineSegment2D to the Rhino scene as a Line."""
    attrib = _get_attributes(layer_name, attributes)
    _line_like_attributes(attrib, line)
    return bake_linesegment2d(line.geometry, z, attributes=attrib)


def bake_display_polygon2d(polygon, z=0, layer_name=None, attributes=None):
    """Add DisplayPolygon2D to the Rhino scene as a Polyline."""
    attrib = _get_attributes(layer_name, attributes)
    _line_like_attributes(attrib, polygon)
    return bake_polygon2d(polygon.geometry, z, attributes=attrib)


def bake_display_arc2d(arc, z=0, layer_name=None, attributes=None):
    """Add DisplayArc2D to the Rhino scene as an Arc or a Circle."""
    attrib = _get_attributes(layer_name, attributes)
    _line_like_attributes(attrib, arc)
    return bake_arc2d(arc.geometry, z, attributes=attrib)


def bake_display_polyline2d(polyline, z=0, layer_name=None, attributes=None):
    """Add DisplayPolyline2D to the Rhino scene as a Curve."""
    attrib = _get_attributes(layer_name, attributes)
    _line_like_attributes(attrib, polyline)
    return bake_polyline2d(polyline.geometry, z, attributes=attrib)


def bake_display_mesh2d(mesh, z=0, layer_name=None, attributes=None):
    """Add DisplayMesh2D to the Rhino scene as a Mesh."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, mesh)
    return bake_mesh2d(mesh.geometry, z, attributes=attrib)


"""____________BAKE 3D DISPLAY GEOMETRY TO THE RHINO SCENE____________"""


def bake_display_vector3d(vector, layer_name=None, attributes=None):
    """Add DisplayVector3D to the Rhino scene as a Line with an Arrowhead."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, vector)
    return bake_vector3d(vector.geometry, attributes=attrib)


def bake_display_point3d(point, layer_name=None, attributes=None):
    """Add ladybug Point3D to the Rhino scene as a Point."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, point)
    return bake_point3d(point.geometry, attributes=attrib)


def bake_display_ray3d(ray, layer_name=None, attributes=None):
    """Add DisplayRay3D to the Rhino scene as a Line with an Arrowhead."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, ray)
    return bake_ray3d(ray.geometry, attributes=attrib)


def bake_display_plane(plane, layer_name=None, attributes=None):
    """Add DisplayPlane to the Rhino scene as a Rectangle."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, plane)
    return bake_plane(plane.geometry, attributes=attrib)


def bake_display_linesegment3d(line, layer_name=None, attributes=None):
    """Add DisplayLineSegment3D to the Rhino scene as a Line."""
    attrib = _get_attributes(layer_name, attributes)
    _line_like_attributes(attrib, line)
    return bake_linesegment3d(line.geometry, attributes=attrib)


def bake_display_arc3d(arc, layer_name=None, attributes=None):
    """Add DisplayArc3D to the Rhino scene as an Arc or a Circle."""
    attrib = _get_attributes(layer_name, attributes)
    _line_like_attributes(attrib, arc)
    return bake_arc3d(arc.geometry, attributes=attrib)


def bake_display_polyline3d(polyline, layer_name=None, attributes=None):
    """Add DisplayPolyline3D to the Rhino scene as a Curve."""
    attrib = _get_attributes(layer_name, attributes)
    _line_like_attributes(attrib, polyline)
    return bake_polyline3d(polyline.geometry, attributes=attrib)


def bake_display_mesh3d(mesh, layer_name=None, attributes=None):
    """Add DisplayMesh3D to the Rhino scene as a Mesh."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, mesh)
    return bake_mesh3d(mesh.geometry, attributes=attrib)


def bake_display_face3d(face, layer_name=None, attributes=None):
    """Add DisplayFace3D to the Rhino scene as a Brep."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, face)
    return bake_face3d(face.geometry, attributes=attrib)


def bake_display_polyface3d(polyface, layer_name=None, attributes=None):
    """Add DisplayPolyface3D to the Rhino scene as a Brep."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, polyface)
    return bake_polyface3d(polyface.geometry, attributes=attrib)


def bake_display_sphere(sphere, layer_name=None, attributes=None):
    """Add DisplaySphere to the Rhino scene as a Brep."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, sphere)
    return bake_sphere(sphere.geometry, attributes=attrib)


def bake_display_cone(cone, layer_name=None, attributes=None):
    """Add DisplayCone to the Rhino scene as a Brep."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, cone)
    return bake_cone(cone.geometry, attributes=attrib)


def bake_display_cylinder(cylinder, layer_name=None, attributes=None):
    """Add DisplayCylinder to the Rhino scene as a Brep."""
    attrib = _get_attributes(layer_name, attributes)
    _color_attribute(attrib, cylinder)
    return bake_cylinder(cylinder.geometry, attributes=attrib)


def bake_display_text3d(display_text, layer_name=None, attributes=None):
    """Add DisplayText3D to the Rhino scene.

    Args:
        display_text: A DisplayText3D object to be added to the Rhino scene.
        layer_name: Optional text string for the layer name on which to place the
            text. If None, text will be added to the current layer.
        attributes: Optional Rhino attributes for adding to the Rhino scene.
    """
    d_txt = display_text.text
    nl_count = len(d_txt.split('\n')) - 1
    if nl_count > 1 and display_text.vertical_alignment == 'Bottom':
        m_vec = display_text.plane.y * (nl_count * display_text.height * -1.5)
        t_pln = display_text.plane.move(m_vec)
    else:
        t_pln = display_text.plane
    txt_h = display_text.height * 0.666
    txt = rd.Text3d(d_txt, from_plane(t_pln), txt_h)
    txt.FontFace = display_text.font
    txt.HorizontalAlignment = TEXT_HORIZ[display_text.horizontal_alignment]
    txt.VerticalAlignment = TEXT_VERT[display_text.vertical_alignment]
    attrib = _get_attributes(layer_name, attributes)
    attrib.ObjectColor = color_to_color(display_text.color)
    return doc.Objects.AddText(txt, attrib)


"""________________EXTRA HELPER FUNCTIONS________________"""


def _color_attribute(attrib, display_obj):
    """Process the attributes of a colored display object."""
    attrib.ColorSource = docobj.ObjectColorSource.ColorFromObject
    attrib.ObjectColor = argb_color_to_color(display_obj.color)


def _line_like_attributes(attrib, display_obj):
    """Process the attributes of a line-like display object."""
    _color_attribute(attrib, display_obj)
    if not isinstance(display_obj.line_width, Default):
        attrib.PlotWeightSource = docobj.ObjectPlotWeightSource.PlotWeightFromObject
        attrib.PlotWeight = display_obj.line_width / LINE_WIDTH_FACTOR
    attrib.LinetypeSource = docobj.ObjectLinetypeSource.LinetypeFromObject
    attrib.LinetypeIndex = LINE_TYPES[display_obj.line_type]



"""Functions to translate from Ladybug geometries to Rhino geometries."""
from .config import tolerance
from .color import color_to_color, gray

try:
    import Rhino.Geometry as rg
except ImportError as e:
    raise ImportError('Failed to import Rhino Geometry.\n{}'.format(e))

try:
    from ladybug_geometry.geometry2d.line import LineSegment2D
except ImportError as e:
    raise ImportError('Failed to import ladybug_geometry.\n{}'.format(e))


"""____________2D GEOMETRY TRANSLATORS____________"""


def from_vector2d(vector):
    """Rhino Vector3d from ladybug Vector2D."""
    return rg.Vector3d(vector.x, vector.y, 0)


def from_point2d(point, z=0):
    """Rhino Point3d from ladybug Point2D."""
    return rg.Point3d(point.x, point.y, z)


def from_ray2d(ray, z=0):
    """Rhino Ray3d from ladybug Ray2D."""
    return rg.Ray3d(from_point2d(ray.p, z), from_vector2d(ray.v))


def from_linesegment2d(line, z=0):
    """Rhino LineCurve from ladybug LineSegment2D."""
    return rg.LineCurve(from_point2d(line.p1, z), from_point2d(line.p2, z))


def from_arc2d(arc, z=0):
    """Rhino Arc from ladybug Arc2D."""
    if arc.is_circle:
        return rg.Circle(from_point2d(arc.c, z), arc.r)
    else:
        pts = (arc.p1, arc.midpoint, arc.p2)
        return rg.Arc(*(from_point2d(pt, z) for pt in pts))


def from_polygon2d(polygon, z=0):
    """Rhino closed PolyLineCurve from ladybug Polygon2D."""
    return rg.PolylineCurve(
        [from_point2d(pt, z) for pt in polygon.vertices] + [from_point2d(polygon[0], z)])


def from_polyline2d(polyline, z=0):
    """Rhino closed PolyLineCurve from ladybug Polyline2D."""
    rhino_pts = [from_point2d(pt, z) for pt in polyline.vertices]
    if polyline.interpolated:
        return rg.Curve.CreateInterpolatedCurve(
            rhino_pts, 3, rg.CurveKnotStyle.UniformPeriodic)
    else:
        return rg.PolylineCurve(rhino_pts)


def from_mesh2d(mesh, z=0):
    """Rhino Mesh from ladybug Mesh2D."""
    pt_function = _get_point2d_function(z)
    return _translate_mesh(mesh, pt_function)


def _get_point2d_function(z_val):
    def point2d_function(pt):
        return from_point2d(pt, z_val)
    return point2d_function


"""____________3D GEOMETRY TRANSLATORS____________"""


def from_vector3d(vector):
    """Rhino Vector3d from ladybug Vector3D."""
    return rg.Vector3d(vector.x, vector.y, vector.z)


def from_point3d(point):
    """Rhino Point3d from ladybug Point3D."""
    return rg.Point3d(point.x, point.y, point.z)


def from_ray3d(ray):
    """Rhino Ray3d from ladybug Ray3D."""
    return rg.Ray3d(from_point3d(ray.p), from_vector3d(ray.v))


def from_linesegment3d(line):
    """Rhino LineCurve from ladybug LineSegment3D."""
    return rg.LineCurve(from_point3d(line.p1), from_point3d(line.p2))


def from_plane(pl):
    """Rhino Plane from ladybug Plane."""
    return rg.Plane(from_point3d(pl.o), from_vector3d(pl.x), from_vector3d(pl.y))


def from_arc3d(arc):
    """Rhino Arc from ladybug Arc3D."""
    if arc.is_circle:
        return rg.Circle(from_plane(arc.plane), from_point3d(arc.c), arc.radius)
    else:
        pts = (arc.p1, arc.midpoint, arc.p2)
        return rg.Arc(*(from_point3d(pt) for pt in pts))


def from_polyline3d(polyline):
    """Rhino closed PolyLineCurve from ladybug Polyline3D."""
    rhino_pts = [from_point3d(pt) for pt in polyline.vertices]
    if polyline.interpolated:
        return rg.Curve.CreateInterpolatedCurve(
            rhino_pts, 3, rg.CurveKnotStyle.UniformPeriodic)
    else:
        return rg.PolylineCurve(rhino_pts)


def from_mesh3d(mesh):
    """Rhino Mesh from ladybug Mesh3D."""
    return _translate_mesh(mesh, from_point3d)


def from_face3d(face):
    """Rhino Brep from ladybug Face3D."""
    segs = [from_linesegment3d(seg) for seg in face.boundary_segments]
    try:
        brep = rg.Brep.CreatePlanarBreps(segs, tolerance)[0]
    except TypeError:  # not planar in Rhino model tolerance; maybe from another model
        print('Brep not planar in Rhino model tolerance. Ignoring tolerance.')
        try:
            brep = rg.Brep.CreatePlanarBreps(segs, 1e6)[0]
        except TypeError:  # it must be a zero-area geometry
            return None
    if face.has_holes:
        for hole in face.hole_segments:
            trim_crvs = [from_linesegment3d(seg) for seg in hole]
            brep.Loops.AddPlanarFaceLoop(0, rg.BrepLoopType.Inner, trim_crvs)
    return brep


def from_polyface3d(polyface):
    """Rhino Brep from ladybug Polyface3D."""
    rh_faces = [from_face3d(face) for face in polyface.faces]
    brep = rg.Brep.JoinBreps(rh_faces, tolerance)
    if len(brep) == 1:
        return brep[0]


def from_sphere(sphere):
    """Rhino Sphere from ladybug Sphere."""
    return rg.Sphere(from_point3d(sphere.center), sphere.radius)


def from_cone(cone):
    """Rhino Cone from ladybug Cone."""
    plane = rg.Plane(from_point3d(cone.vertex), from_vector3d(cone.axis.normalize()))
    return rg.Cone(plane, cone.height, cone.radius)


def from_cylinder(cylinder):
    """Rhino Cone from ladybug Cone."""
    return rg.Cylinder(from_arc3d(cylinder.base_bottom), cylinder.height)


"""________ADDITIONAL 3D GEOMETRY TRANSLATORS________"""


def from_polyline2d_to_joined_polyline(polylines, z=0):
    """Rhino PolylineCurve made by joining list of Polyline2D.

    Args:
        polylines: An array of Ladybug Polyline2D or LineSegment2D to be joined
            into a polyline.
        z: A number for the Z-coordinate. (Default: 0).

    Returns:
        A Rhino brep constructed from the polygon and the offset distance. If offset
        is unsuccessful, just the polyline will be returned.
    """
    line_crv = []
    for pl in polylines:
        if isinstance(pl, LineSegment2D):
            line_crv.append(from_linesegment2d(pl))
        else:
            line_crv.append(from_polyline2d(pl))
    return rg.Curve.JoinCurves(line_crv)[0]


def from_polyline2d_to_offset_brep(polylines, offset, z=0):
    """Rhino Brep made by offsetting a joined list of Polyline2D inward.

    Args:
        polylines: An array of Ladybug Polyline2D or LineSegment2D to be joined
            and translated to an offset Brep.
        offset: A number for the distance to offset the Polygon inward.
        z: A number for the Z-coordinate. (Default: 0).

    Returns:
        A Rhino brep constructed from the polygon and the offset distance. If offset
        is unsuccessful, just the polyline will be returned.
    """
    curve = from_polyline2d_to_joined_polyline(polylines, z)
    crv_style = rg.CurveOffsetCornerStyle.Sharp
    all_curves = [curve]
    off_curves = curve.Offset(rg.Plane.WorldXY, -offset, tolerance, crv_style)
    if off_curves is not None:
        all_curves.extend(off_curves)
        offset_brep = rg.Brep.CreatePlanarBreps(all_curves)
        if len(offset_brep) == 1:
            if offset_brep[0].Loops.Count == 2:
                return offset_brep[0]
    return curve


def from_face3d_to_wireframe(face):
    """Rhino PolyLineCurves from ladybug Face3D.

    Args:
        face: A Ladybug Face3D object to be translated to a wireframe.

    Returns:
        A list of Rhino polyline curves for the boundary and holes in the face.
    """
    boundary = [_polyline_points(face.boundary)]
    if face.has_holes:
        return boundary + [_polyline_points(tup) for tup in face.holes]
    return boundary


def from_polyface3d_to_wireframe(polyface):
    """Rhino PolyLineCurve from ladybug Polyface3D."""
    return [f for face in polyface.faces for f in from_face3d_to_wireframe(face)]


def from_face3d_to_solid(face, offset):
    """Rhino Solid Brep from a ladybug Face3D and an offset from the base face.

    Args:
        face: Ladybug geometry Face3D object.
        offset: Number for the offset distance from the base face.
    """
    srf_brep = from_face3d(face)
    return rg.Brep.CreateFromOffsetFace(
        srf_brep.Faces[0], offset, tolerance, False, True)


def from_face3ds_to_joined_brep(faces):
    """A list of joined Breps from an array of ladybug Face3D."""
    return rg.Brep.JoinBreps([from_face3d(face) for face in faces], tolerance)


def from_face3ds_to_colored_mesh(faces, color):
    """Colored Rhino mesh from an array of ladybug Face3D and ladybug Color.

    This is used in workflows such as coloring Model geometry with results.
    """
    joined_mesh = rg.Mesh()
    for face in faces:
        try:
            joined_mesh.Append(from_mesh3d(face.triangulated_mesh3d))
        except Exception:
            pass  # failed to create a Rhino Mesh from the Face3D
    joined_mesh.VertexColors.CreateMonotoneMesh(color_to_color(color))
    return joined_mesh


def from_mesh2d_to_outline(mesh, z=0):
    """Rhino Polylines from the faces of ladybug Mesh2D."""
    pt_function = _get_point2d_function(z)
    verts = [pt_function(pt) for pt in mesh.vertices]
    face_plines = []
    for face in mesh.faces:
        outline = [verts[f] for f in face] + [verts[face[0]]]
        face_plines.append(rg.PolylineCurve(outline))
    return face_plines


def from_mesh3d_to_outline(mesh):
    """Rhino Mesh and Polylines from the ladybug Mesh3D.

    Returns:
        A tuple with two items - a Rhino Mesh first followed by outline curves
        of the Mesh.
    """
    rh_mesh = from_mesh3d(mesh)
    return rh_mesh, rh_mesh.GetNakedEdges()


"""________________EXTRA HELPER FUNCTIONS________________"""


def _translate_mesh(mesh, pt_function):
    """Translates both 2D and 3D meshes to Rhino"""
    rhino_mesh = rg.Mesh()
    if mesh.is_color_by_face:  # Mesh is constructed face-by-face
        _f_num = 0
        for face in mesh.faces:
            for pt in tuple(mesh[i] for i in face):
                rhino_mesh.Vertices.Add(pt_function(pt))
            if len(face) == 4:
                rhino_mesh.Faces.AddFace(
                    _f_num, _f_num + 1, _f_num + 2, _f_num + 3)
                _f_num += 4
            else:
                rhino_mesh.Faces.AddFace(_f_num, _f_num + 1, _f_num + 2)
                _f_num += 3
        if mesh.colors is not None:
            rhino_mesh.VertexColors.CreateMonotoneMesh(gray())
            _f_num = 0
            for i, face in enumerate(mesh.faces):
                col = color_to_color(mesh.colors[i])
                rhino_mesh.VertexColors[_f_num] = col
                rhino_mesh.VertexColors[_f_num + 1] = col
                rhino_mesh.VertexColors[_f_num + 2] = col
                if len(face) == 4:
                    rhino_mesh.VertexColors[_f_num + 3] = col
                    _f_num += 4
                else:
                    _f_num += 3
    else:  # Mesh is constructed vertex-by-vertex
        for pt in mesh.vertices:
            rhino_mesh.Vertices.Add(pt_function(pt))
        for face in mesh.faces:
            rhino_mesh.Faces.AddFace(*face)
        if mesh.colors is not None:
            rhino_mesh.VertexColors.CreateMonotoneMesh(gray())
            for i, col in enumerate(mesh.colors):
                rhino_mesh.VertexColors[i] = color_to_color(col)
    return rhino_mesh


def _polyline_points(tup):
    """Convert a tuple of Ladybug Geometry points to a Rhino Polyline."""
    return rg.PolylineCurve([from_point3d(pt) for pt in tup] + [from_point3d(tup[0])])



"""Functions for getting viewport properties, creating new viewports, and editing them.
"""
import math

try:
    import System
except ImportError as e:  # No .NET; We are really screwed
    raise ImportError("Failed to import System.\n{}".format(e))

try:
    import Rhino.Geometry as rg
    import Rhino.Display as rd
    import Rhino.RhinoDoc as rhdoc
    import Rhino.ApplicationSettings.AppearanceSettings as aps
except ImportError as e:  # No RhinoCommon doc is available. This module is useless.
    raise ImportError("Failed to import Rhino.\n{}".format(e))

try:
    import scriptcontext as sc
except ImportError as e:  # No Rhino doc is available. This module is useless.
    raise ImportError("Failed to import Rhino scriptcontext.\n{}".format(e))

from .text import TextGoo


def camera_oriented_plane(origin):
    """Get a Rhino Plane that is oriented facing the camera.

    Args:
        origin: A Rhino Point for the origin of the plane.
    """
    active_view = sc.doc.Views.ActiveView.ActiveViewport
    camera_x = active_view.CameraX
    camera_y = active_view.CameraY
    return rg.Plane(origin, camera_x, camera_y)


def orient_to_camera(geometry, position=None):
    """Orient an array of Rhino geometry objects to the camera of the active viewport.

    Args:
        geometry: An array of Rhino Geometry objects (or TextGoo objects) to
            the camera of the active Rhino viewport.
        position: A point to be used as the origin around which the the geometry
            will be oriented. If None, the lower left corner of the bounding box
            around the geometry will be used.
    """

    # set the default position if it is None
    origin = _bounding_box_origin(geometry)
    pt = origin if position is None else position

    # get a plane oriented to the camera
    oriented_plane = camera_oriented_plane(pt)

    # orient the input geometry to the plane facing the camera
    base_plane = rg.Plane(origin, rg.Vector3d(0, 0, 1))
    xform = rg.Transform.PlaneToPlane(base_plane, oriented_plane)
    geo = []
    for rh_geo in geometry:
        if isinstance(rh_geo, TextGoo):
            geo.append(rh_geo.Transform(xform))
        elif isinstance(rh_geo, rg.Point3d):
            geo.append(oriented_plane)
        elif isinstance(rh_geo, rg.Plane):
            geo.append(oriented_plane)
        else:
            new_geo = rh_geo.Duplicate()
            new_geo.Transform(xform)
            geo.append(new_geo)
    return geo


def viewport_by_name(view_name=None):
    """Get a Rhino Viewport object using the name of the viewport.

    Args:
        view_name: Text for the name of the Rhino Viewport. If None, the
            current Rhino viewport will be used. If the view is a named view that
            is not currently open, it will be restored to the active view of
            the Rhino document.
    """
    try:
        return sc.doc.Views.Find(view_name, False).ActiveViewport \
            if view_name is not None else sc.doc.Views.ActiveView.ActiveViewport
    except Exception:
        # try to find a named view and restore it
        view_table = rhdoc.ActiveDoc.NamedViews
        for i, viewp in enumerate(view_table):
            if viewp.Name == view_name:
                active_viewp = sc.doc.Views.ActiveView.ActiveViewport
                view_table.Restore(i, active_viewp)
                return active_viewp
        else:
            raise ValueError('Viewport "{}" was not found in the Rhino '
                             'document.'.format(view_name))


def open_viewport(view_name, width=None, height=None):
    """Create a new Viewport in the active Rhino document at specified dimensions.

    This will also set the newly-created view to be tha active Viewport.

    Args:
        view_name: Text for the name of the new Rhino Viewport that will be created.
        width: Optional positive integer for the width of the view in pixels. If None,
            the width of the currently active viewport will be used.
        height: Optional positive integer for the height of the view in pixels. If
            None, the height of the currently active viewport will be used.
    """
    # close the view if it already exists
    if sc.doc.Views.Find(view_name, False):
        sc.doc.Views.Find(view_name, False).Close()

    # get the width and the height if it was not specified
    w = sc.doc.Views.ActiveView.ActiveViewport.Size.Width if not width else width
    h = sc.doc.Views.ActiveView.ActiveViewport.Size.Height if not height else height

    # compute the X,Y screen coordinates where the new viewport will be placed
    x = round((System.Windows.Forms.Screen.PrimaryScreen.Bounds.Width - w) / 2)
    y = round((System.Windows.Forms.Screen.PrimaryScreen.Bounds.Height - h) / 2)
    rec = System.Drawing.Rectangle(System.Drawing.Point(x, y), System.Drawing.Size(w, h))

    # add the new view to the rhino document
    rhdoc.ActiveDoc.Views.Add(
        view_name, rd.DefinedViewportProjection.Perspective, rec, True)
    return viewport_by_name(view_name)


def set_view_display_mode(viewport, display_mode):
    """Set the display mode of a Rhino Viewport.

    Args:
        viewport: A Rhino ViewPort object, which will have its display mode set.
        display_mode: Text for the display mode to which the Rhino viewport will be
            set. For example: Wireframe, Shaded, Rendered, etc.
    """
    mode_obj = rd.DisplayModeDescription.FindByName(display_mode)
    viewport.DisplayMode = mode_obj


def set_iso_view_direction(viewport, direction, center_point=None):
    """Set a Rhino Viewport to have an isometric view in a specific direction.

    Args:
        viewport: A Rhino ViewPort object, which will have its direction set.
        direction: A Rhino vector that will be used to set the direction of
            the isometric view.
        center_point: Optional Rhino point for the target of the camera. If no point
            is provided, the Rhino origin will be used (0, 0, 0).
    """
    viewport.ChangeToParallelProjection(True)
    center_point = center_point if center_point is not None else rg.Point3d.Origin
    viewport.SetCameraLocation(rg.Point3d.Add(center_point, direction), False)
    viewport.SetCameraTarget(center_point, False)
    viewport.SetCameraDirection(direction, False)


def capture_view(viewport, file_path, width=None, height=None, display_mode=None,
                 transparent=False):
    """Capture a Viewport to a PNG file path.

    Args:
        viewport: A Rhino ViewPort object, which will have its display mode set.
        file_path: Full path to the file where the image will be saved.
        width: Integer for the image width in pixels. If None, the width of the
            active viewport will be used. (Default: None).
        height: Integer for the image height in pixels. If None, the height of the
            active viewport will be used. (Default: None).
        display_mode: Text for the display mode to which the Rhino viewport will be
            set. For example: Wireframe, Shaded, Rendered, etc. If None, it will
            be the current viewport's display mode. (Default: None).
        transparent: Boolean to note whether the background of the image should be
            transparent or have the same color as the Rhino scene. (Default: False).

    Returns:
        Full path to the image file that was written.
    """
    # create the view capture object
    active_viewp = viewport_by_name()
    img_w = active_viewp.Size.Width if width is None else width
    img_h = active_viewp.Size.Height if height is None else height
    img_size = System.Drawing.Size(img_w, img_h)

    # capture the view
    if display_mode is not None:
        mode_obj = rd.DisplayModeDescription.FindByName(display_mode)
        pic = viewport.ParentView.CaptureToBitmap(img_size, mode_obj)
    else:
        pic = viewport.ParentView.CaptureToBitmap(img_size)

    # remove the background color if requested
    if transparent:
        back_col = aps.ViewportBackgroundColor
        if (display_mode is None and viewport.DisplayMode.EnglishName == 'Rendered') \
                or display_mode == 'Rendered':
            back_col = sc.doc.Views.Document.RenderSettings.BackgroundColorTop
        pic.MakeTransparent(back_col)

    # save the bitmap to a png file
    if not file_path.endswith('.png'):
        file_path = '{}.png'.format(file_path)
    System.Drawing.Bitmap.Save(pic, file_path)
    return file_path


def viewport_vh_vv(viewport, view_type):
    """Get the horizontal angle (vh) and the vertical angle (vv) from a viewport.

    Args:
        viewport: A Rhino ViewPort object for which properties will be extracted.
        view_type: An integer to set the view type (-vt). Choose from the
            choices below.

            * 0 Perspective (v)
            * 1 Hemispherical fisheye (h)
            * 2 Parallel (l)
            * 3 Cylindrical panorama (c)
            * 4 Angular fisheye (a)
            * 5 Planisphere [stereographic] projection (s)

    """
    if view_type == 0:  # perspective
        right_vec = viewport.GetFrustumRightPlane()[1][1]
        left_vec = viewport.GetFrustumLeftPlane()[1][1]
        h_angle = 180 - math.degrees(rg.Vector3d.VectorAngle(right_vec, left_vec))
        bottom_vec = viewport.GetFrustumBottomPlane()[1][1]
        top_vec = viewport.GetFrustumTopPlane()[1][1]
        v_angle = 180 - math.degrees(rg.Vector3d.VectorAngle(bottom_vec, top_vec))
        return h_angle, v_angle
    if view_type == 1 or view_type == 5:
        return 180, 180
    if view_type == 2:
        v_rect = viewport.GetNearRect()
        return int(v_rect[0].DistanceTo(v_rect[1])), int(v_rect[0].DistanceTo(v_rect[2]))
    if view_type == 3:
        return 360, 180
    if view_type == 4:
        return 60, 60


def viewport_properties(viewport, view_type=None):
    """Get a dictionary of properties of a Rhino viewport.

    Args:
        viewport: A Rhino ViewPort object for which properties will be extracted.
        view_type: An integer to set the view type (-vt). Choose from the
            choices below or set to None to have it derived from the viewport.

            * 0 Perspective (v)
            * 1 Hemispherical fisheye (h)
            * 2 Parallel (l)
            * 3 Cylindrical panorama (c)
            * 4 Angular fisheye (a)
            * 5 Planisphere [stereographic] projection (s)

    Returns:
        A dictionary with the following keys: 'view_type', 'position', 'direction',
        'up_vector', 'h_angle', 'v_angle'
    """
    # ensure that we have an integer for the view_type
    if view_type is None:
        view_type = 2 if viewport.IsParallelProjection else 0

    # get the position, direction and up vectors
    pos = viewport.CameraLocation
    direct = viewport.CameraDirection
    up_vec = viewport.CameraUp
    direct.Unitize()
    up_vec.Unitize()

    # get the h_angle and v_angle from the viewport
    h_angle, v_angle = viewport_vh_vv(viewport, view_type)

    return {
        'view_type': view_type,
        'position': (pos.X, pos.Y, pos.Z),
        'direction': (direct.X, direct.Y, direct.Z),
        'up_vector': (up_vec.X, up_vec.Y, up_vec.Z),
        'h_angle': h_angle,
        'v_angle': v_angle
    }


def _bounding_box_origin(geometry):
    """Get the origin of a bounding box around a list of geometry.

    Args:
        geometry: A list of geometry for which the bounding box origin will
            be computed.
    """
    # get the first geometry
    first_geo = geometry[0]
    # if the geometry is a point or plane, just return the plane origin
    if isinstance(first_geo, rg.Point3d):
        return first_geo
    elif isinstance(first_geo, rg.Plane):
        return first_geo.Origin
    # assume that the geometry is a bunch of breps, meshes or text objects
    b_box = first_geo.GetBoundingBox(False) if not isinstance(first_geo, TextGoo) \
        else first_geo.get_Boundingbox()
    for geo in geometry[1:]:
        if isinstance(geo, TextGoo):
            b_box = rg.BoundingBox.Union(b_box, geo.get_Boundingbox())
        else:
            b_box = rg.BoundingBox.Union(b_box, geo.GetBoundingBox(False))
    return b_box.Corner(True, True, True)



"""Functions for managing user resources like standards and measures."""
import os

try:
    from ladybug.futil import preparedir, nukedir
except ImportError as e:
    raise ImportError("Failed to import ladybug.\n{}".format(e))

STANDARDS_SUBFOLDERS = (
    'constructions', 'constructionsets', 'schedules', 'programtypes',
    'modifiers', 'modifiersets'
)


def setup_resource_folders(overwrite=False):
    """Set up user resource folders in their respective locations.

    Args:
        overwrite: Boolean to note whether the user resources should only be set
            up if they do not exist, in which case existing resources will be
            preserved, or should they be overwritten.
    """
    # first check if there's an environment variable available for APPDATA
    app_folder = os.getenv('APPDATA')
    if app_folder is not None:
        resource_folder = os.path.join(app_folder, 'ladybug_tools')
        # set up user standards
        lib_folder = os.path.join(resource_folder, 'standards')
        for sub_f in STANDARDS_SUBFOLDERS:
            sub_lib_folder = os.path.join(lib_folder, sub_f)
            if not os.path.isdir(sub_lib_folder) or overwrite:
                preparedir(sub_lib_folder)
        # set up the user weather
        epw_folder = os.path.join(resource_folder, 'weather')
        if not os.path.isdir(epw_folder) or overwrite:
            if os.path.isdir(epw_folder):
                nukedir(epw_folder, rmdir=True)  # delete all sub-folders
            preparedir(epw_folder)
        # set up the user measures folder
        measure_folder = os.path.join(resource_folder, 'measures')
        if not os.path.isdir(measure_folder) or overwrite:
            if os.path.isdir(measure_folder):
                nukedir(measure_folder, rmdir=True)  # delete all sub-folders
            preparedir(measure_folder)
        return resource_folder



"""Functions for creating user objects."""
import os

try:
    import Grasshopper.Folders as Folders
    import Grasshopper.Kernel as gh
except ImportError:
    raise ImportError("Failed to import Grasshopper.")

try:
    from ladybug.config import folders
except ImportError as e:
    raise ImportError("Failed to import ladybug.\n{}".format(e))

# find the location where the Grasshopper user objects are stored
UO_FOLDER = Folders.UserObjectFolders[0]
GHA_FOLDER = Folders.DefaultAssemblyFolder
if os.name == 'nt':
    # search all assembly folders to see if they live in the core installation
    lbt_components = os.path.join(folders.ladybug_tools_folder, 'grasshopper')
    if os.path.isdir(lbt_components):
        comp_dir = 'C:\\ProgramData\\McNeel\\Rhinoceros\\packages'
        for a_fold in Folders.AssemblyFolders:
            a_fold = str(a_fold)
            if a_fold.startswith(comp_dir) and 'LadybugTools' in a_fold:
                # a special plugin loader has been added
                UO_FOLDER = lbt_components
                GHA_FOLDER = lbt_components
                break

# map from the AdditionalHelpFromDocStrings to values for user object exposure
EXPOSURE_MAP = (
    gh.GH_Exposure.dropdown,
    gh.GH_Exposure.primary,
    gh.GH_Exposure.secondary,
    gh.GH_Exposure.tertiary,
    gh.GH_Exposure.quarternary,
    gh.GH_Exposure.quinary,
    gh.GH_Exposure.senary,
    gh.GH_Exposure.septenary
)

# map from the component category to the plugin package folder
FOLDER_MAP = {
    'Ladybug': 'ladybug_grasshopper',
    'Honeybee': 'honeybee_grasshopper_core',
    'HB-Radiance': 'honeybee_grasshopper_radiance',
    'HB-Energy': 'honeybee_grasshopper_energy',
    'Dragonfly': 'dragonfly_grasshopper',
    'LB-Legacy': 'LB-Legacy',
    'HB-Legacy': 'HB-Legacy',
    'HoneybeePlus': 'HoneybeePlus'
}


def create_userobject(component, move=True):
    """Create UserObject from a component.

    Args:
        component: A Grasshopper Python component.
        move: A Boolean to note whether the component should be moved to a subdirectory
            based on FOLDER_MAP. (Default: True).
    """
    # initiate userobject
    uo = gh.GH_UserObject()
    # set attributes
    uo.Icon = component.Icon_24x24
    try:
        exposure = int(component.AdditionalHelpFromDocStrings)
    except Exception:  # no exposure category specified
        exposure = 1
    uo.Exposure = EXPOSURE_MAP[exposure]
    uo.BaseGuid = component.ComponentGuid
    uo.Description.Name = component.Name
    uo.Description.Description = component.Description
    uo.Description.Category = component.Category
    uo.Description.SubCategory = component.SubCategory

    # save the userobject to a file
    uo.SetDataFromObject(component)
    uo.SaveToFile()

    # move the user object file to the assigned folder
    if move:
        ufd = os.path.join(UO_FOLDER, FOLDER_MAP[component.Category], 'user_objects')
        ufp = os.path.join(ufd, '%s.ghuser' % component.Name)
        if not os.path.isdir(ufd):
            # create folder if it is not already created
            os.mkdir(ufd)
        elif os.path.isfile(ufp):
            # remove current userobject
            try:
                os.remove(ufp)
            except Exception:
                pass  # access is denied to the user object location
        uo.Path = ufp

    uo.SaveToFile()
    return uo



"""Functions for managing the setting of Rhino's IronPython path."""
import os
import io
import xml.etree.ElementTree

try:
    from ladybug.futil import nukedir, copy_file_tree
    from ladybug.config import folders as lb_folders
except ImportError as e:
    raise ImportError("Failed to import ladybug.\n{}".format(e))


# core library packages, which get copied or cleaned out of the Rhino scripts folder
PACKAGES = (
    'ladybug_rhino', 'ladybug_geometry', 'ladybug_geometry_polyskel',
    'ladybug', 'ladybug_comfort', 'honeybee', 'honeybee_standards', 'honeybee_energy',
    'honeybee_radiance', 'honeybee_radiance_folder', 'honeybee_radiance_command',
    'dragonfly', 'dragonfly_energy', 'dragonfly_radiance', 'dragonfly_uwg',
    'lbt_recipes', 'pollination_handlers'
)
# Rhino versions that the plugins are compatible with
RHINO_VERSIONS = ('6.0', '7.0')
# UUID that McNeel uses to identify the IronPython plugin
IRONPYTHON_ID = '814d908a-e25c-493d-97e9-ee3861957f49'


def create_python_package_dir():
    """Get the default path where the ladybug_tools Python packages are installed.

    This method works both on Windows and Mac. If the folder is not found, this
    method will create the folder.
    """
    py_install = os.path.join(lb_folders.ladybug_tools_folder, 'python')
    py_path = os.path.join(py_install, 'Lib', 'site-packages') if os.name == 'nt' \
        else os.path.join(py_install, 'lib', 'python3.7', 'site-packages')
    if not os.path.isdir(py_path):
        return os.makedirs(py_path)
    return py_path


def iron_python_search_path(python_package_dir, settings_file=None,
                            destination_file=None):
    """Set Rhino to search for libraries in a given directory (on either OS).

    This is used as part of the installation process to ensure that Grasshopper
    looks for the core Python libraries in the ladybug_tools folder. The file
    will not be edited if the python_package_dir is already in the settings file.

    Args:
        python_package_dir: The path to a directory that contains the Ladybug
            Tools core libraries.
        settings_file: An optional XML settings file to which the python_package_dir
            will be added. If None, this method will search the current user's
            folder for all copies of this file for the installed Rhino versions.
        destination_file: Optional destination file to write out the edited settings
            file. If it is None, the settings_file will be overwritten.
    """
    # make sure that the rhino scripts folder is clean to avoid namespace issues
    clean_rhino_scripts()
    # set the plugin to look for packages in ladybug_tools folder
    new_settings = []
    if os.name == 'nt':  # we are on Windows
        all_settings = [settings_file] if settings_file is not None else \
            find_ironpython_settings_windows()
        for sf in all_settings:
            dest_settings = iron_python_search_path_windows(
                python_package_dir, sf, destination_file)
            new_settings.append(dest_settings)
    else:  # we are on Mac, Linux, or some other unix-based system
        copy_packages_to_rhino_scripts(python_package_dir)
    return new_settings


def find_installed_rhino_versions_windows():
    """Get a list of the compatible Rhino versions installed on this Windows machine."""
    program_files = os.getenv('PROGRAMFILES')
    installed_vers = []
    for ver in RHINO_VERSIONS:
        rhino_path = os.path.join(
            program_files, 'Rhino {}'.format(ver[0]), 'System', 'Rhino.exe')
        if os.path.isfile(rhino_path):
            installed_vers.append(ver)
    return installed_vers


def find_ironpython_settings_windows():
    """Get a list of all settings XML files for the supported RHINO_VERSIONS."""
    installed_set_files = []
    appdata_roaming = os.getenv('APPDATA')
    for ver in find_installed_rhino_versions_windows():
        # get the settings folder or create it if it doesn't exist
        plugin_folder = os.path.join(
            appdata_roaming, 'McNeel', 'Rhinoceros', ver, 'Plug-ins')
        settings_path = os.path.join(
            plugin_folder, 'IronPython ({})'.format(IRONPYTHON_ID), 'settings')
        if not os.path.isdir(settings_path):
            os.makedirs(settings_path)
        # append the default settings to the list of files to edit
        sf = os.path.join(settings_path, 'settings-Scheme__Default.xml')
        installed_set_files.append(sf)
        # get the search paths for any Rhino-inside instances if they exist
        for set_file in os.listdir(settings_path):
            if set_file.startswith('settings-Scheme') and \
                    set_file != 'settings-Scheme__Default.xml':
                sf = os.path.join(settings_path, set_file)
                installed_set_files.append(sf)
    return installed_set_files


def iron_python_search_path_windows(python_package_dir, settings_file,
                                    destination_file=None):
    """Set Rhino to search for libraries in a given directory (on Windows).

    This is used as part of the installation process to ensure that Grasshopper
    looks for the core Python libraries in the ladybug_tools folder. The file
    will not be edited if the python_package_dir is already in the settings file.

    Args:
        python_package_dir: The path to a directory that contains the Ladybug
            Tools core libraries.
        settings_file: An XML settings file to which the python_package_dir
            will be added.
        destination_file: Optional destination file to write out the edited settings
            file. If it is None, the settings_file will be overwritten.
    """
    # open the settings file and find the search paths
    search_path_needed = True
    settings_key_needed = False
    existing_paths = None
    if os.path.isfile(settings_file):
        with io.open(settings_file, 'r', encoding='utf-8') as fp:
            set_data = fp.read()
        element = xml.etree.ElementTree.fromstring(set_data)
        settings = element.find('settings')
        if settings is not None:
            for entry in settings.iter('entry'):
                if 'SearchPaths' in list(entry.attrib.values()):
                    existing_paths = entry.text
        else:  # there's no settings key within the XML file; we must add it
            search_path_needed = False
            settings_key_needed = True
    else:
        contents = [
            '<?xml version="1.0" encoding="utf-8"?>',
            '<settings id="2.0">',
            '<settings>',
            '</settings>',
            '</settings>'
        ]
        with open(settings_file, 'w') as fp:
            fp.write('\n'.join(contents))

    # add the search paths if it was not found
    if destination_file is None:
        destination_file = settings_file
    if search_path_needed:
        if 'ladybug_tools' in python_package_dir and existing_paths is not None:
            existing_paths = filter_existing_paths(existing_paths)
        new_paths = '{};{}'.format(existing_paths, python_package_dir) \
            if existing_paths is not None and existing_paths != '' \
            else python_package_dir
        line_to_add = '    <entry key="SearchPaths">{}</entry>\n'.format(new_paths)
        with io.open(settings_file, 'r', encoding='utf-8') as fp:
            contents = fp.readlines()
        line_to_del = None
        for i, line in enumerate(contents):
            if '<entry key="SearchPaths">' in line:
                line_to_del = i
            elif '</settings>' in line:
                break
        contents.insert(i, line_to_add)
        if line_to_del is not None:
            del contents[line_to_del]
        with io.open(destination_file, 'w', encoding='utf-8') as fp:
            fp.write(''.join(contents))
    elif settings_key_needed:
        lines_to_add = '  <settings>\n    <entry key="SearchPaths">{}</entry>\n' \
            '  </settings>\n'.format(python_package_dir)
        with io.open(settings_file, 'r', encoding='utf-8') as fp:
            contents = fp.readlines()
        for i, line in enumerate(contents):
            if '</settings>' in line:
                break
        contents.insert(i, lines_to_add)
        with io.open(destination_file, 'w', encoding='utf-8') as fp:
            fp.write(''.join(contents))
    return destination_file


def filter_existing_paths(existing_paths):
    """Filter out any duplicate/unwanted search paths."""
    paths_list = existing_paths.split(';')
    filt_paths = [p for p in paths_list
                  if 'ladybug_tools' not in p and 'pollination' not in p]
    return ';'.join(filt_paths)


def find_installed_rhino_scripts():
    """Get the path to the current user's Rhino scripts folder if it exists."""
    installed_scripts = []
    if os.name == 'nt':  # we are on Windows
        appdata_roaming = os.getenv('APPDATA')
        for ver in find_installed_rhino_versions_windows():
            scripts_folder = os.path.join(appdata_roaming, 'McNeel',
                                          'Rhinoceros', ver, 'scripts')
            if os.path.isdir(scripts_folder):
                installed_scripts.append(scripts_folder)
    else:  # we are on Mac
        home_folder = os.getenv('HOME') or os.path.expanduser('~')
        for ver in RHINO_VERSIONS:
            scripts_folder = os.path.join(home_folder, 'Library', 'Application Support',
                                          'McNeel', 'Rhinoceros', ver, 'scripts')
            if not os.path.isdir(scripts_folder):
                os.makedirs(scripts_folder)
            installed_scripts.append(scripts_folder)
    return installed_scripts


def copy_packages_to_rhino_scripts(python_package_dir, directory=None):
    """Copy Ladybug tools packages into a directory.

    Args:
        python_package_dir: The path to a directory that contains the Ladybug
            Tools core libraries.
        directory: The directory into which the packages will be copied. If None,
            the function will search for all installed copies of the current user's
            Rhino scripts folder.
    """
    folders = find_installed_rhino_scripts() if directory is None else [directory]
    for fold in folders:
        for pkg in PACKAGES:
            lib_folder = os.path.join(python_package_dir, pkg)
            dest_folder = os.path.join(fold, pkg)
            if os.path.isdir(lib_folder):
                copy_file_tree(lib_folder, dest_folder, True)
                print('Python packages copied to: {}'.format(dest_folder))


def clean_rhino_scripts(directory=None):
    """Remove installed Ladybug Tools packages from the old library directory.

    This function is usually run in order to avoid potential namespace conflicts.

    Args:
        directory: The directory to be cleaned. If None, the function will
            search for all installed copies of the current user's
            Rhino scripts folder.
     """
    folders = find_installed_rhino_scripts() if directory is None else [directory]
    for fold in folders:
        for pkg in PACKAGES:
            lib_folder = os.path.join(fold, pkg)
            if os.path.isdir(lib_folder):
                nukedir(lib_folder, True)
                print('Python packages removed from: {}'.format(lib_folder))



# coding=utf-8
from ladybug_rhino.pythonpath import iron_python_search_path_windows

import os
import io
import plistlib


def test_iron_python_search_path_windows():
    """Test the iron_python_search_path_windows method with a sample XML file."""
    package_dir = os.path.join('ladybug_tools', 'python', 'Lib', 'site-packages')
    settings_file = './tests/ironpython/settings-Scheme__Default.xml'
    dest_file = './tests/ironpython/new_settings.xml'

    edited_file = iron_python_search_path_windows(package_dir, settings_file, dest_file)
    assert os.path.isfile(edited_file)

    with io.open(edited_file, 'r', encoding='utf-8') as fp:
        set_data = fp.read()
    assert package_dir in set_data
    os.remove(edited_file)


def test_iron_python_search_path_windows_mingbo():
    """Test the iron_python_search_path_windows method with Mingbo's sample XML file."""
    package_dir = os.path.join('ladybug_tools', 'python', 'Lib', 'site-packages')
    settings_file = './tests/ironpython/settings-Scheme__Default_mingbo.xml'
    dest_file = './tests/ironpython/new_settings_mingbo.xml'

    edited_file = iron_python_search_path_windows(package_dir, settings_file, dest_file)
    assert os.path.isfile(edited_file)

    with io.open(edited_file, 'r', encoding='utf-8') as fp:
        set_data = fp.read()
    assert package_dir in set_data
    os.remove(edited_file)



"""Functions to bake entire Ladybug objects into the Rhino scene.

The methods here are intended to help translate groups of geometry that are
commonly generated by several objects in Ladybug core (ie. legends, compasses,
visualization sets, etc.)
"""
import json

from ladybug_geometry.geometry2d import Vector2D, Point2D, Ray2D, LineSegment2D, \
    Arc2D, Polyline2D, Polygon2D, Mesh2D
from ladybug_geometry.geometry3d import Vector3D, Point3D, Ray3D, LineSegment3D, \
    Arc3D, Polyline3D, Plane, Mesh3D, Face3D, Polyface3D, Sphere, Cone, Cylinder
from ladybug.graphic import GraphicContainer
from ladybug_display.geometry2d import DisplayVector2D, DisplayPoint2D, \
    DisplayRay2D, DisplayLineSegment2D, DisplayPolyline2D, DisplayArc2D, \
    DisplayPolygon2D, DisplayMesh2D
from ladybug_display.geometry3d import DisplayVector3D, DisplayPoint3D, \
    DisplayRay3D, DisplayPlane, DisplayLineSegment3D, DisplayPolyline3D, DisplayArc3D, \
    DisplayFace3D, DisplayMesh3D, DisplayPolyface3D, DisplaySphere, DisplayCone, \
    DisplayCylinder, DisplayText3D
from ladybug_display.visualization import AnalysisGeometry

from .config import units_system
from .color import color_to_color
from .bakegeometry import bake_point2d, bake_vector2d, bake_ray2d, \
    bake_linesegment2d, bake_arc2d, bake_polygon2d, bake_polyline2d, bake_mesh2d, \
    bake_point3d, bake_vector3d, bake_ray3d, bake_linesegment3d, bake_plane, \
    bake_arc3d, bake_polyline3d, bake_mesh3d, bake_face3d, bake_polyface3d, \
    bake_sphere, bake_cone, bake_cylinder, _get_layer, _get_attributes
from .bakedisplay import bake_display_point2d, bake_display_vector2d, \
    bake_display_ray2d, bake_display_linesegment2d, bake_display_arc2d, \
    bake_display_polygon2d, bake_display_polyline2d, bake_display_mesh2d, \
    bake_display_text3d, bake_display_vector3d, bake_display_point3d, \
    bake_display_ray3d, bake_display_linesegment3d, bake_display_arc3d, \
    bake_display_polyline3d, bake_display_plane, bake_display_mesh3d, \
    bake_display_face3d, bake_display_polyface3d, bake_display_sphere, \
    bake_display_cone, bake_display_cylinder

try:
    import System
except ImportError as e:
    raise ImportError("Failed to import Windows/.NET libraries\n{}".format(e))

try:
    import Rhino.DocObjects as docobj
    import Rhino.RhinoDoc as rhdoc
    doc = rhdoc.ActiveDoc
except ImportError as e:
    raise ImportError("Failed to import Rhino document attributes.\n{}".format(e))

try:
    import Grasshopper as gh
except ImportError:
    print('Failed to import Grasshopper.Grasshopper Baking disabled.')
    gh = None

BAKE_MAPPER = {
    Vector2D: bake_vector2d,
    Point2D: bake_point2d,
    Ray2D: bake_ray2d,
    LineSegment2D: bake_linesegment2d,
    Arc2D: bake_arc2d,
    Polyline2D: bake_polyline2d,
    Polygon2D: bake_polygon2d,
    Mesh2D: bake_mesh2d,
    Vector3D: bake_vector3d,
    Point3D: bake_point3d,
    Ray3D: bake_ray3d,
    LineSegment3D: bake_linesegment3d,
    Arc3D: bake_arc3d,
    Polyline3D: bake_polyline3d,
    Plane: bake_plane,
    Mesh3D: bake_mesh3d,
    Face3D: bake_face3d,
    Polyface3D: bake_polyface3d,
    Sphere: bake_sphere,
    Cone: bake_cone,
    Cylinder: bake_cylinder,
    DisplayVector2D: bake_display_vector2d,
    DisplayPoint2D: bake_display_point2d,
    DisplayRay2D: bake_display_ray2d,
    DisplayLineSegment2D: bake_display_linesegment2d,
    DisplayPolyline2D: bake_display_polyline2d,
    DisplayArc2D: bake_display_arc2d,
    DisplayPolygon2D: bake_display_polygon2d,
    DisplayMesh2D: bake_display_mesh2d,
    DisplayVector3D: bake_display_vector3d,
    DisplayPoint3D: bake_display_point3d,
    DisplayRay3D: bake_display_ray3d,
    DisplayPlane: bake_display_plane,
    DisplayLineSegment3D: bake_display_linesegment3d,
    DisplayPolyline3D: bake_display_polyline3d,
    DisplayArc3D: bake_display_arc3d,
    DisplayFace3D: bake_display_face3d,
    DisplayMesh3D: bake_display_mesh3d,
    DisplayPolyface3D: bake_display_polyface3d,
    DisplaySphere: bake_display_sphere,
    DisplayCone: bake_display_cone,
    DisplayCylinder: bake_display_cylinder,
    DisplayText3D: bake_display_text3d
}


def bake_legend(legend, layer_name=None):
    """Add a Ladybug Legend object to the Rhino scene.

    Args:
        legend: A Ladybug Legend object to be added to the Rhino scene.
        layer_name: Optional text string for the layer name on which to place the
            legend. If None, text will be added to the current layer.

    Returns:
        A list of IDs that point to the objects in the Rhino scene in the following
        order:

        -   legend_mesh -- A colored mesh for the legend.

        -   legend_title -- A text object for the legend title.

        -   legend_text -- Text objects for the rest of the legend text.
    """
    # bake the legend mesh
    legend_mesh = bake_mesh3d(legend.segment_mesh, layer_name)
    # translate the legend text
    _height = legend.legend_parameters.text_height
    _font = legend.legend_parameters.font
    if legend.legend_parameters.continuous_legend is False:
        legend_text = [
            DisplayText3D(txt, loc, _height, None, _font, 'Left', 'Bottom')
            for txt, loc in zip(legend.segment_text, legend.segment_text_location)]
    elif legend.legend_parameters.vertical is True:
        legend_text = [
            DisplayText3D(txt, loc, _height, None, _font, 'Left', 'Center')
            for txt, loc in zip(legend.segment_text, legend.segment_text_location)]
    else:
        legend_text = [
            DisplayText3D(txt, loc, _height, None, _font, 'Center', 'Bottom')
            for txt, loc in zip(legend.segment_text, legend.segment_text_location)]
    legend_title = DisplayText3D(
        legend.title, legend.title_location, _height, None, _font)
    legend_text.insert(0, legend_title)
    # bake the text objects
    legend_text_guids = []
    for txt_obj in legend_text:
        legend_text_guids.append(bake_display_text3d(txt_obj, layer_name))
    return [legend_mesh] + legend_text_guids


def bake_analysis(analysis, layer_name=None, bake_3d_legend=False,
                  min_point=None, max_point=None):
    """Add a Ladybug Display AnalysisGeometry object to the Rhino scene.

    Args:
        analysis: A Ladybug Display AnalysisGeometry object to be added to
            the Rhino scene.
        layer_name: Optional text string for the parent layer name on which to
            place the AnalysisGeometry. The actual layer of the context will
            always have a name that aligns with the AnalysisGeometry.display_name.
        bake_3d_legend: A Boolean to note whether the AnalysisGeometry should
            be baked with 3D legends for any AnalysisGeometries it
            includes. (Default: False).
        min_point: An optional Point3D to override the default min point
            that are used to generate the legend. (Default: None).
        max_point: An optional Point3D to override the default max point
            that are used to generate the legend. (Default: None).

    Returns:
        A list of IDs that point to the objects in the Rhino scene.
    """
    # get attributes corresponding to the layer
    layer_name = analysis.display_name if layer_name is None else \
        '{}::{}'.format(layer_name, analysis.display_name)
    min_pt = analysis.min_point if min_point is None else min_point
    max_pt = analysis.max_point if max_point is None else max_point
    # generate the colors that correspond to the values
    obj_ids = []
    for i, data in enumerate(analysis.data_sets):
        # get properties used for all analysis geometries
        objs_to_group = []
        graphic = GraphicContainer(
            data.values, min_pt, max_pt,
            data.legend_parameters, data.data_type, data.unit)
        colors = graphic.value_colors
        sub_layer_name = layer_name if data.data_type is None else \
            '{}::{}'.format(layer_name, data.data_type.name)
        layer_index = _get_layer(sub_layer_name)
        # translate the analysis geometry using the matching method
        if analysis.matching_method == 'faces':
            c_count = 0
            for mesh in analysis.geometry:
                mesh.colors = colors[c_count:c_count + len(mesh.faces)]
                c_count += len(mesh.faces)
                bake_func = bake_mesh3d if isinstance(mesh, Mesh3D) else bake_mesh2d
                objs_to_group.append(bake_func(mesh, layer_name=layer_index))
        elif analysis.matching_method == 'vertices':
            c_count = 0
            for mesh in analysis.geometry:
                mesh.colors = colors[c_count:c_count + len(mesh.vertices)]
                c_count += len(mesh.vertices)
                bake_func = bake_mesh3d if isinstance(mesh, Mesh3D) else bake_mesh2d
                objs_to_group.append(bake_func(mesh, layer_name=layer_index))
        else:  # one color per geometry object
            bake_func = BAKE_MAPPER[analysis.geometry[0].__class__]
            for geo_obj, col in zip(analysis.geometry, colors):
                attrib = _get_attributes(layer_index)
                attrib.ColorSource = docobj.ObjectColorSource.ColorFromObject
                attrib.ObjectColor = color_to_color(col)
                objs_to_group.append(bake_func(geo_obj, attributes=attrib))
        # group the objects, and add JSON of values to layer user data
        group_table = doc.Groups  # group table
        group_table.Add(sub_layer_name, objs_to_group)
        layer_table = doc.Layers  # layer table
        layer_obj = layer_table[layer_index]
        layer_obj.UserDictionary.Set('vis_data', json.dumps(data.to_dict()))
        layer_obj.UserDictionary.Set('guids', System.Array[System.Guid](objs_to_group))
        if i != analysis.active_data:  # hide the inactive data layer
            layer_obj.IsVisible = False
        # add geometry to the global list and bake the legend if requested
        obj_ids.extend(objs_to_group)
        if bake_3d_legend:
            obj_ids.extend(bake_legend(graphic.legend, layer_index))
    # hide the layer if it is hidden
    if analysis.hidden:
        layer_table = doc.Layers  # layer table
        layer_index = _get_layer(layer_name)
        layer_obj = layer_table[layer_index]
        layer_obj.IsVisible = False
    return obj_ids


def bake_context(context, layer_name=None):
    """Add a Ladybug Display ContextGeometry object to the Rhino scene.

    Args:
        context: A Ladybug Display ContextGeometry object to be added to
            the Rhino scene.
        layer_name: Optional text string for the parent layer name on which to
            place the ContextGeometry. The actual layer of the context will
            always have a name that aligns with the ContextGeometry.display_name.

    Returns:
        A list of IDs that point to the objects in the Rhino scene.
    """
    # get attributes corresponding to the layer
    layer_name = context.display_name if layer_name is None else \
        '{}::{}'.format(layer_name, context.display_name)
    layer_index = _get_layer(layer_name)
    # hide the layer if it is hidden
    if context.hidden:
        layer_table = doc.Layers  # layer table
        layer_obj = layer_table[layer_index]
        layer_obj.IsVisible = False
    # loop through the objects and add them to the scene
    obj_ids = []
    for geo_obj in context.geometry:
        bake_func = BAKE_MAPPER[geo_obj.__class__]
        obj_ids.append(bake_func(geo_obj, layer_index))
    return obj_ids


def bake_visualization_set(vis_set, bake_3d_legend=False):
    """Add a Ladybug Display VisualizationSet object to the Rhino scene.

    Args:
        context_geometry: A Ladybug VisualizationSet object to be added to
            the Rhino scene.
        bake_3d_legend: A Boolean to note whether the VisualizationSet should
            be baked with 3D legends for any AnalysisGeometries it
            includes. (Default: False).

    Returns:
        A list of IDs that point to the objects in the Rhino scene.
    """
    # convert the visualization set to model units if necessary
    units_sys = units_system()
    if vis_set.units is not None and vis_set.units != units_sys:
        vis_set.convert_to_units(units_sys)
    # bake all of the geometries
    obj_ids = []
    for geo in vis_set.geometry:
        if isinstance(geo, AnalysisGeometry):  # translate it as AnalysisGeometry
            a_objs = bake_analysis(
                geo, vis_set.display_name, bake_3d_legend,
                vis_set.min_point, vis_set.max_point)
            obj_ids.extend(a_objs)
        else:  # translate it as ContextGeometry
            obj_ids.extend(bake_context(geo, vis_set.display_name))
    return obj_ids


class VisSetGoo(gh.Kernel.IGH_BakeAwareData):
    """A Bake-able version of the VisualizationSet for Grasshopper.

    Args:
        visualization_set: A Ladybug Display VisualizationSet object to be bake-able
            in the Rhino scene.
    """

    def __init__(self, visualization_set):
        self.vis_set = visualization_set

    def BakeGeometry(self, doc, att, id):
        try:
            if self.vis_set is not None:
                guids = bake_visualization_set(self.vis_set, True)
                return True, guids
        except Exception as e:
            System.Windows.Forms.MessageBox.Show(str(e), 'script error')
            return False, System.Guid.Empty

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        return str(self.vis_set)



"""Functions for managing the copying of user objects to the Grasshopper path."""
import os

try:
    from ladybug.futil import nukedir, copy_file_tree
except ImportError as e:
    raise ImportError("Failed to import ladybug.\n{}".format(e))

from .pythonpath import RHINO_VERSIONS

# core library packages, which get copied or cleaned out of the Rhino scripts folder
PACKAGES = \
    ('ladybug_grasshopper', 'honeybee_grasshopper_core', 'honeybee_grasshopper_energy',
     'honeybee_grasshopper_radiance', 'dragonfly_grasshopper')
# package containing .gha files
DOTNET_PACKAGES = ('ladybug_grasshopper_dotnet',)
GRASSHOPPER_ID = 'b45a29b1-4343-4035-989e-044e8580d9cf'


def copy_components_packages(directory):
    """Copy all Ladybug tools components packages to their respective locations.

    Args:
        directory: The path to a directory that contains all of the Ladybug
            Tools Grasshopper python packages to be copied (both user object
            packages and dotnet gha packages).
    """
    clean_userobjects()
    copy_packages_to_userobjects(directory)
    clean_libraries()
    copy_packages_to_libraries(directory)


def find_grasshopper_userobjects():
    """Get the paths to the current user's Grasshopper user object folder.

    The folder(s) will be created if they do not already exist.
    """
    if os.name == 'nt':  # we are on Windows
        appdata_roaming = os.getenv('APPDATA')
        uo_folder = [os.path.join(appdata_roaming, 'Grasshopper', 'UserObjects')]
    else:  # we are on Mac
        home_folder = os.getenv('HOME') or os.path.expanduser('~')
        uo_folder = []
        for ver in RHINO_VERSIONS:
            uo_fold = os.path.join(
                home_folder, 'Library', 'Application Support', 'McNeel',
                'Rhinoceros', ver, 'Plug-ins', 'Grasshopper ({})'.format(GRASSHOPPER_ID),
                'UserObjects')
            uo_folder.append(uo_fold)
    for uo_fold in uo_folder:
        if not os.path.isdir(uo_fold):
            os.makedirs(uo_fold)
    return uo_folder


def copy_packages_to_userobjects(directory):
    """Copy Ladybug Tools user object packages to the current user's userobject folder.

    Args:
        directory: The path to a directory that contains the Ladybug
            Tools Grasshopper python packages to be copied.
    """
    uo_folders = find_grasshopper_userobjects()
    for uo_folder in uo_folders:
        for pkg in PACKAGES:
            lib_folder = os.path.join(directory, pkg)
            dest_folder = os.path.join(uo_folder, pkg)
            if os.path.isdir(lib_folder):
                copy_file_tree(lib_folder, dest_folder, True)
                print('UserObjects copied to: {}'.format(dest_folder))


def clean_userobjects():
    """Remove installed Ladybug Tools packages from the user's userobjects folder."""
    uo_folders = find_grasshopper_userobjects()
    for uo_folder in uo_folders:
        for pkg in PACKAGES:
            lib_folder = os.path.join(uo_folder, pkg)
            if os.path.isdir(lib_folder):
                nukedir(lib_folder, True)
                print('UserObjects removed from: {}'.format(lib_folder))


def find_grasshopper_libraries():
    """Get the paths to the current user's Grasshopper Libraries folder.

    The folder(s) will be created if they do not already exist.
    """
    if os.name == 'nt':  # we are on Windows
        appdata_roaming = os.getenv('APPDATA')
        lib_folder = [os.path.join(appdata_roaming, 'Grasshopper', 'Libraries')]
    else:  # we are on Mac
        home_folder = os.getenv('HOME') or os.path.expanduser('~')
        lib_folder = []
        for ver in RHINO_VERSIONS:
            lib_fold = os.path.join(
                home_folder, 'Library', 'Application Support', 'McNeel',
                'Rhinoceros', ver, 'Plug-ins', 'Grasshopper ({})'.format(GRASSHOPPER_ID),
                'Libraries')
            lib_folder.append(lib_fold)
    for lib_fold in lib_folder:
        if not os.path.isdir(lib_fold):
            os.makedirs(lib_fold)
    return lib_folder


def copy_packages_to_libraries(directory):
    """Copy Ladybug tools Libraries packages to the current user's libraries folder.

    Args:
        directory: The path to a directory that contains the Ladybug
            Tools Grasshopper python packages.
    """
    lib_folders = find_grasshopper_libraries()
    for lib_folder in lib_folders:
        for pkg in DOTNET_PACKAGES:
            src_folder = os.path.join(directory, pkg)
            dest_folder = os.path.join(lib_folder, pkg)
            if os.path.isdir(src_folder):
                copy_file_tree(src_folder, dest_folder, True)
                print('Components copied to: {}'.format(dest_folder))


def clean_libraries():
    """Remove installed Ladybug Tools packages from the user's Libraries folder."""
    lib_folders = find_grasshopper_libraries()
    for lib_folder in lib_folders:
        for pkg in DOTNET_PACKAGES:
            lib_folder = os.path.join(lib_folder, pkg)
            if os.path.isdir(lib_folder):
                nukedir(lib_folder, True)
                print('Components removed from: {}'.format(lib_folder))
